#include "typedef.h"

#include <assert.h>
#include <time.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "object.h"
#include "xmalloc.h"
#include "util.h"
#include "opcode.h"
#include "code.h"
#include "file.h"
#include "source.h"
#include "class.h"
#include "function.h"
#include "detect.h"
#include "clics_output.h"

#ifndef NDEBUG
//#define DEBUG_DETECT
//#define DEBUG_HILLCLIMB
//#define TEST_HILLCLIMB
#define PROOF
#endif

extern int		g_hypertext_cnt;
extern int		g_top_clones;
extern int		g_top_cols;
extern int		g_report_source;
extern int		g_report_assembler;
extern int		g_verbose;
extern char	const *g_html_dirP;
extern int		g_min_match_sequence;
extern int		g_min_function_sequence;
extern int		g_mismatch_cost;
extern int		g_same_cost;
extern int		g_match_lines;
extern int  	g_comparisons;
extern int		g_hillclimb;
extern int		g_hillclimb_timeout;
extern int		g_iterations;
extern int		g_best_weight;
extern time_t	g_hillclimb_elapse;

extern long long	g_original_matches;
extern long long	g_extra_matches;
extern long long	g_cancelled_matches;
extern int			g_max_blocks_seen;
extern int			g_max_iterations_seen;

static	Csource			*source1P;		// Source 1st side looking at
static	Csource			*source2P;		// Source 2nd side looking at
static	Cclass			*class1P;		// Class 1st side looking at
static	Cclass			*class2P;		// Class 2nd side looking at
static	Cfunction		*function1P;	// Function 1st side looking at
static	Cfunction		*function2P;	// Function 2nd side looking at
static	codeT			*function_code1P;// Start of function1 code
static	codeT			*function_code2P;// Start of function2 code
static	codeT			*function_end1P;// Beyond end of function1 code
static	codeT			*function_end2P;// Beyond end of function2 code
static	codeT			*start_code1P;	// Comparing from 1st side
static	codeT			*start_code2P;	// Comparing from 2nd side
static	codeT			*end_code1P;	// May not compare to here on 1st side
static	codeT			*end_code2P;	// May not compare to here on 2nd side
static	codeT			*best_end1P;	// Best end of side 1
static	codeT			*best_end2P;	// Best end of side 2
static  int				forward_jumps;	// Matched forward jumps in clone

extern FILE		*stdoutF;
extern FILE		*redirectF;

static Cclass	*last_classP  = 0;
static FILE		*sourceF      = 0;
static FILE		*cloneF       = 0;

typedef struct {
	unsigned int	weight;
	unsigned int	html;
} topT;

static topT		*top_clonesP       = 0;
static topT		*top_clones_endP   = 0;
static int		top_clones_cnt     = 0;

/*
void
trap(void)
{
}
*/

void
initialise_detect(void)
{
	int	need;

	if (!g_html_dirP || (!g_report_assembler && !g_report_source)) {
		g_top_clones = 0;
	}
	if (g_top_clones > 0) {
		need = g_top_clones * sizeof(topT);
		top_clonesP = (topT *) Xmalloc(need);
		if (!top_clonesP) {
			outofmemory();
		}
		memset(top_clonesP, 0, need);
		top_clones_endP = (topT *) (((char *) top_clonesP) + need);
	}
}

void
start_report(void)
{
	clics_sources();
	fputs("<p>\n<table>\n", redirectF);
}

static int		g_lth1, g_lth2;
static time_t	start_clock, end_clock;

#ifdef PROOF
static void
proof(void)
{
	static	codeT	*code1aP, *code2aP, *code1bP, *code2bP;
	static	codeT	*code1a_jumpP, *code2a_jumpP;
	static	codeT	*code1b_jumpP, *code2b_jumpP;

	assert(start_code1P->m_matchP == start_code2P);
	assert(start_code2P->m_matchP == start_code1P);
	
	for (code1aP = start_code1P; code1aP < best_end1P; ++code1aP) {
		if (!(code2aP = code1aP->m_matchP)) {
			continue;
		}
		assert(code2aP->m_matchP == code1aP);
		code1a_jumpP = code1aP->m_jump.P;
		code2a_jumpP = code2aP->m_jump.P;
		if (!code1a_jumpP) {
			assert(!code2a_jumpP);
		} else if (code1a_jumpP == code1aP) {
			assert(code2a_jumpP == code2aP);
		} else if (code1a_jumpP > code1aP) {
			assert(code2a_jumpP > code2aP);
		} else {
			assert(code2a_jumpP && code2a_jumpP < code2aP);
			assert(code1a_jumpP >= start_code1P);
			assert(code2a_jumpP >= start_code2P);
		}
		for (code1bP = start_code1P; code1bP < best_end1P; ++code1bP) {
			if (code1aP == code1bP) {
				continue;
			}
			if (!(code2bP = code1bP->m_matchP)) {
				continue;
			}
			assert((code1aP > code1bP && code2aP > code2bP) ||
                   (code1aP < code1bP && code2aP < code2bP));
			if (!code1a_jumpP) {
				continue;
			}
			assert((code1a_jumpP <= code1bP && code2a_jumpP <= code2bP) ||
				   (code1a_jumpP > code1bP  && code2a_jumpP > code2bP));

			code1b_jumpP = code1bP->m_jump.P;
			code2b_jumpP = code2bP->m_jump.P;
			if (!code1b_jumpP || !code2b_jumpP) {
				continue;
			}
			if (code1a_jumpP >= best_end1P && code2a_jumpP >= best_end2P) {
				continue;
			}
			if (code1b_jumpP >= best_end1P && code2b_jumpP >= best_end2P) {
				continue;
			}
			assert((code1a_jumpP >= code1b_jumpP && code2a_jumpP >= code2b_jumpP) ||
				   (code1a_jumpP <= code1b_jumpP && code2a_jumpP <= code2b_jumpP));
	}	}	
}
#endif

extern FILE	*clicsF;

static void
dump_clics_clone(codeT *start_codeP, codeT *best_endP)
{
	Cfile		*fileP;
	char		**sourcePP;
	positionT	*position1P, *position2P;
	int			clone_start_lineno, clone_start_offset, clone_end_lineno, clone_end_column, clone_end_offset;
	char		*start_sourceP, *last_charP;

	clone_start_lineno = 0;
	clone_start_offset = 0;
	clone_end_lineno   = 0;
	clone_end_column   = 0;
	clone_end_offset   = 0;

	position1P = start_codeP->m_positionP;
	if (position1P) {
		fileP = position1P->source_fileP;
		if (fileP) {
			position2P = best_endP->m_positionP;
			if (position2P && position2P->source_fileP == fileP) {
				clone_start_lineno = position1P->line - 1;
				clone_end_lineno   = position2P->line - 1;
				sourcePP = fileP->m_sourcePP;
				if (!sourcePP) {
					fileP->load_source();
					sourcePP = fileP->m_sourcePP;
				}
				if (sourcePP != (char **) -1) {
					start_sourceP      = sourcePP[0];
					assert(0 <= clone_start_lineno && clone_start_lineno < fileP->m_source_lines);
					clone_start_offset = sourcePP[clone_start_lineno] - start_sourceP;
					assert(0 < clone_end_lineno && clone_end_lineno <= fileP->m_source_lines);
					clone_end_offset   = sourcePP[clone_end_lineno]   - start_sourceP;
					last_charP         = sourcePP[clone_end_lineno] - 1;
					clone_end_offset   = last_charP - start_sourceP;
					--clone_end_lineno;
					clone_end_column   = last_charP - sourcePP[clone_end_lineno]; 
	}	}	}	}

	fprintf(clicsF, "%d\t%d,0,%d\t%d,%d,%d\t", 
		fileP->m_number, 
		clone_start_lineno, 
		clone_start_offset,
		clone_end_lineno,
		clone_end_column,
		clone_end_offset);
}

static void
report_clone(void)
{
	extern	FILE			*taF;
	extern	FILE			*tempF;

	static	int				lineno1, lineno2, next_lineno1, next_lineno2;
	static	int				different1, different2;

	static	u2T				*source_name1P, *source_name2P;
	static	codeT			*at_code1P, *at_code2P;
	static	positionT		*position1P, *position2P;
	static	Cfile			*file1P, *file2P, *next_file1P, *next_file2P;

	static	int				offset1, offset2;
	static	int				report_source;
	static	unsigned int	hypertext_cnt;

	static	unsigned int heap_index;
	static	topT	*heapP, *leftP, *rightP;
	static	unsigned int	weight;
	static	char			*P;
	static time_t			clock_now;

	source_name1P = class1P->m_nameP->P;
	source_name2P = class2P->m_nameP->P;

	// Increment the detail page number
	g_hypertext_cnt += 1;

	if (taF) {
		Cclone *clone1P, *clone2P;

		clone1P = function1P->locate_clone(start_code1P, best_end1P);
		clone2P = function2P->locate_clone(start_code2P, best_end2P);

		fprintf(taF, "R C%p C%p\n", clone1P, clone2P);
		if (g_html_dirP && (g_report_source || g_report_assembler)) {
			fprintf(tempF, "(R C%p C%p) { html=%u }\n", clone1P, clone2P, g_hypertext_cnt);
	}	}

	if (clicsF) {
		dump_clics_clone(start_code1P, best_end1P);
		dump_clics_clone(start_code2P, best_end2P);
		fprintf(clicsF, "%ld\n", (long) (best_end1P - start_code1P + best_end2P - start_code2P));
	}

	if (top_clonesP) {
		// Use a heap with the smallest weight at the top

		weight = g_lth1 + g_lth2;
		if (top_clonesP->weight < weight) {
			++top_clones_cnt;
			heap_index = 1;
			heapP      = top_clonesP;
			// Add entry to the heap
			for (;;) {
				leftP = heapP + heap_index;
				if (top_clones_endP <= leftP) {
					goto found_heap_entry;
				}
				rightP = leftP + 1;
				if ((top_clones_endP <= rightP) || (leftP->weight <= rightP->weight)) {
					if (weight <= leftP->weight) {
						goto found_heap_entry;
					}
					*heapP       = *leftP;
					heapP        = leftP;
					heap_index <<= 1;
					continue;
				}
				if (weight <= rightP->weight) {
found_heap_entry:	heapP->weight = weight;
					heapP->html   = g_hypertext_cnt;
					break;
				}
				*heapP       = *rightP;
				heapP        = rightP;
				heap_index   = (heap_index << 1) + 1;
			}
	}	}

	if (g_html_dirP) {
		if (last_classP != class1P) {
			if (sourceF) {
				fprintf(sourceF,"</table>\n<p>\nEnd of report</body>\n</html>\n");
				fclose(sourceF);
			}
			hypertext_cnt = class1P->m_hypertext_cnt;
			sprintf(g_filename,"%s/%u.html", g_html_dirP, hypertext_cnt);
			sourceF = open_file(g_filename);
			last_classP = class1P;

			stdoutF       = sourceF;
			
			fputs( HTML_HEADER "<title>Clones in ", stdoutF);
			Cfunction::dump_descriptor(source_name1P);
			fputs("</title>\n<body>\n<a href=index.html>Top</a>\n<h3>Clones in ", stdoutF);
			Cfunction::dump_descriptor(source_name1P);
			fputs("</h3>\n<table>\n", stdoutF);

			// On top level page write
			stdoutF   = redirectF;
			clock_now = time(0);
			P         = ctime(&clock_now);
			P[19]     = 0;
			fprintf(stdoutF, "<tr><td>%s</td><td><a href=\"%u.html\">", P + 11, hypertext_cnt);
			Cfunction::dump_descriptor(source_name1P);
			fprintf(stdoutF, "</a></td></tr>\n");
		}

		cloneF = 0;
		if (g_report_source || g_report_assembler) {
			sprintf(g_filename,"%s/%u.html", g_html_dirP, g_hypertext_cnt);
			cloneF = open_file(g_filename);
			fputs(
HTML_HEADER
"<script language=\"javascript\">\n"
"<!--\n"
"var a, s;\n"
"var a_shown='none'\n"
"var s_shown='none'\n"
"function toggleA() {\n"
"  for (i = 0; i < a.length; i++) { a[i].style.display = a_shown}\n"
"  a_shown = (a_shown == 'none' ? 'inline' : 'none')\n"
"}\n"
"function toggleS() {\n"
"  for (i = 0; i < s.length; i++) { s[i].style.display = s_shown}\n"
"  s_shown = (s_shown == 'none' ? 'inline' : 'none')\n"
"}\n"
"function init(){ \n"
"  var ai = 0;\n"
"  var si = 0;\n"
"  var tr = document.all.tags(\"tr\");\n"
"  var e, className, i;\n"
"  for (i = 0; i < tr.length; i++) {\n"
"    e = tr[i];\n"
"    className = \"\" + e.className;\n"
"    switch (className) {\n"
"      case 'a': ++ai; break;\n"
"      case 's': ++si; break;\n"
"  } }\n"
"  a  = new Array(ai);\n"
"  s  = new Array(si);\n"
"  ai = 0;\n"
"  si = 0;\n"
"  for (i = 0; i < tr.length; i++) {\n"
"    e = tr[i];\n"
"    className = \"\" + e.className;\n"
"    switch(className) {\n"
"    case 'a': a[ai] = e; ++ai; break;\n"
"    case 's': s[si] = e; ++si; break;\n"
"} } }\n"
"//-->\n"
"</script>\n"
"</head>\n"
"<body onLoad=\"init()\">\n"
"<p>\n"
"[<a href=index.html>Top</a>] [<a href=\"javascript:toggleA()\">Toggle</a>] Pcode\n"
"[<a href=\"javascript:toggleS()\">Toggle</a>] Source\n"
"<p>\n"
				  , cloneF);
		}
			
		stdoutF = sourceF;
		fputs("<tr><td><i>", stdoutF);
		fprintf(stdoutF, "<a href=\"%u.html\">", g_hypertext_cnt);
		function1P->dump_descriptor();
		fputs("</a>", stdoutF);
		fprintf(stdoutF, "</i> <font color=\"red\">[%d]</font></td>\n    <td><i>", (int) (best_end1P - start_code1P));
		function2P->dump_descriptor();
		fprintf(stdoutF, "</i> <font color=\"blue\">[%d]</font></td>\n", (int) (best_end2P - start_code2P));

		fputs("</i></td>\n</tr>\n", stdoutF);
		stdoutF = cloneF;
	} 
	fputs("<p>\n<table>\n", stdoutF);
	fprintf(stdoutF, "<tr><td><b><a href=%u.html>", class1P->m_hypertext_cnt);
	Cfunction::dump_descriptor(source_name1P);
	fprintf(stdoutF, "</a></b></td>\n    <td><b><a href=%u.html>", class2P->m_hypertext_cnt);
	Cfunction::dump_descriptor(source_name2P);
	fputs("</a></b></td>\n</tr>\n<tr><td><i>", stdoutF);
	function1P->dump_descriptor();
	fprintf(stdoutF, "</i> <font color=\"red\">[%d]</font></td>\n    <td><i>", (int) (best_end1P - start_code1P));
	function2P->dump_descriptor();
	fprintf(stdoutF, "</i> <font color=\"blue\">[%d]</font></td>\n</tr>\n", (int) (best_end2P - start_code2P));

	if (g_report_source || g_report_assembler) {

		fputs("<tr><td><hr></hr></td><td><hr></hr></td></tr>", stdoutF); 

		at_code1P = start_code1P;
		at_code2P = start_code2P;

		if ((report_source = g_report_source)) {
			report_source = 0;
			different1    = 0;
			different2    = 0;

			for (at_code1P = best_end1P; at_code1P > start_code1P; ) {
				--at_code1P;
				position1P = at_code1P->m_positionP;
				if (position1P) {
					file1P = position1P->source_fileP;
					if (!file1P->m_sourcePP) {
						file1P->load_source();
					}
					if (file1P->m_source_lines) {
						report_source = 1;
			}	}	}
			for (at_code2P = best_end2P; at_code2P > start_code2P; ) {
				--at_code2P;
				position2P = at_code2P->m_positionP;
				if (position2P) {
					file2P = position2P->source_fileP;
					if (!file2P->m_sourcePP) {
						file2P->load_source();
					}
					if (file2P->m_source_lines) {
						report_source = 1;
			}	}	}	
			next_lineno1 = -1;
			next_lineno2 = -1;
			if ((position1P = start_code1P->m_positionP) && (file1P = position1P->source_fileP) && file1P->m_source_lines) {
				next_file1P  = file1P;
				next_lineno1 = file1P->m_source_lines;
			}
			if ((position2P = start_code2P->m_positionP) && (file2P = position2P->source_fileP) && file2P->m_source_lines) {
				next_file2P  = file2P;
				next_lineno2 = file2P->m_source_lines;
		}	}


		offset1 = at_code1P - function_code1P;
		offset2 = at_code2P - function_code2P;
		for (; at_code1P < best_end1P || at_code2P < best_end2P; ) {

			if (g_report_assembler) {
				if (best_end1P <= at_code1P || (at_code1P->m_matchP && at_code1P->m_matchP != at_code2P) ) {
					fprintf(stdoutF, "<tr class=a><td></td>\n    <td><font color=\"blue\">%d: ", offset2);
					Cfunction::dump_pcode(at_code2P);
					fputs("</font></td>\n</tr>\n", stdoutF);
					++at_code2P;
					++offset2;
					different2 = 1;
				} else if (!at_code1P->m_matchP) {
					fprintf(stdoutF, "<tr class=a><td><font color=\"red\">%d: ", offset1);
					Cfunction::dump_pcode(at_code1P);
					fputs("</font></td>\n    <td>", stdoutF);
					++at_code1P;
					++offset1;
					different1 = 1;
				
					if (at_code2P < best_end2P && !at_code2P->m_matchP) {
						fprintf(stdoutF, "<font color=\"blue\">%d: ", offset2);
						Cfunction::dump_pcode(at_code2P);
						fputs("</font>", stdoutF);
						++at_code2P;
						++offset2;
						different2 = 1;
					}
					fputs("</td>\n</tr>\n", stdoutF);
				} else {
					fputs("<tr class=a><td>", stdoutF);
					Cfunction::dump_pcode(at_code1P);
					fputs("</td>\n    <td>", stdoutF);
					Cfunction::dump_pcode(at_code2P);
					fputs("</td>\n</tr>\n", stdoutF);

					++at_code1P;
					++offset1;
					++at_code2P;
					++offset2;
				} 
			} else {
				// Not reporting assembler

				if (best_end1P <= at_code1P || (at_code1P->m_matchP && at_code1P->m_matchP != at_code2P) ) {
					++at_code2P;
					different2 = 1;
				} else if (!at_code1P->m_matchP) {
					++at_code1P;
					different1 = 1;
				
					if (at_code2P < best_end2P && !at_code2P->m_matchP) {
						++at_code2P;
					}
				} else {
					++at_code1P;
					++at_code2P;
			}	} 

			if (report_source) {

				lineno1      = next_lineno1;
				lineno2      = next_lineno2;
				file1P       = next_file1P;
				file2P       = next_file2P;

				next_lineno1 = -1;
				next_file1P  = 0;
				if (at_code1P < best_end1P) {
					position1P = at_code1P->m_positionP;
					if (position1P) {
						next_file1P = position1P->source_fileP;
						if (next_file1P->m_source_lines) {
							next_lineno1 = position1P->line;
						} else {
							next_file1P = 0;
				}	}	} 

				next_lineno2 = -1;
				next_file2P  = 0;
				if (at_code2P < best_end2P) {
					position2P = at_code2P->m_positionP;
					if (position2P) {
						next_file2P = position2P->source_fileP;
						if (next_file2P->m_source_lines) {
							next_lineno2 = position2P->line;
						} else {
							next_file2P = 0;
				}	}	} 

				if (lineno1 == next_lineno1 && file1P == next_file1P) {
					if (!file2P || lineno2 == next_lineno2) {
						continue;
					}
					fputs("<tr class=s><td></td>\n    <td><font color=\"blue\">", stdoutF);
					file2P->showline(lineno2);
					fputs("</font></td>\n</tr>\n", stdoutF);
				} else if (lineno2 == next_lineno2) {
					if (!file1P || lineno1 == next_lineno1) {
						continue;
					}
					fputs("<tr class=s><td><font color=\"red\">", stdoutF);
					file1P->showline(lineno1);
					fputs("</font></td>\n    <td></td></tr>\n", stdoutF);
				} else {
					fputs("<tr class=s><td>", stdoutF);
					if (file1P) {
						if (!different1) {
							file1P->showline(lineno1);
						} else {
							fputs("<font color=\"red\">", stdoutF);
							file1P->showline(lineno1);
							fputs("</font>", stdoutF);
					}	}
					fputs("</td><td>", stdoutF);
					if (file2P) {
						if (!different2) {
							file2P->showline(lineno2);
						} else {
							fputs("<font color=\"blue\">", stdoutF);
							file2P->showline(lineno2);
							fputs("</font>", stdoutF);
					}	}
					fputs("</td></tr>\n", stdoutF);
				}
				different1 = 0;
				different2 = 0;
			}
		}
	}
	fputs("</table>\n", stdoutF);
	if (g_html_dirP) {
		fprintf(stdoutF, "<p><table>\n<tr><td>Start</td><td>%s</td></tr>\n", 	ctime(&start_clock));
		fprintf(stdoutF, "<tr><td>End</td><td>%s</td></tr>\n</table>\n", 	ctime(&end_clock));
		fputs("<p>End of report\n</body>\n</html>\n", stdoutF);
		fclose(stdoutF);
		stdoutF = redirectF;
	}
}

void
finalize_report(void)
{
    extern const char *g_signalP;

	static	topT	*topP, *atP, *endP, *heapP, *leftP, *rightP;
	static	int		heap_index, cnt, i;
	static	FILE	*topF;

	if (sourceF) {
		fprintf(sourceF,"</table>\n");
        if (g_signalP) {
            fprintf(sourceF, "<p><font color=red>**EXECUTION INTERRUPTED BY SIG%s**</font>\n", g_signalP);
        }
		fprintf(sourceF,"<p>\n<p>End of report</body>\n</html>\n");
		fclose(sourceF);
	}
	
	if (top_clonesP) {
		cnt = top_clones_cnt;
		if (cnt > g_top_clones) {
			cnt = g_top_clones;
		}
		if (cnt) {
			topP = atP = (topT *) Xmalloc(cnt *sizeof(topT));
			if (!topP) {
				outofmemory();
			}
			for (endP = atP + cnt; ; ) {
				if (top_clonesP->html) {
					*atP++     = *top_clonesP;
					if (endP <= atP) {
						break;
				}	}
				heapP      = top_clonesP;
				heap_index = 1;
				for (;;) {
					leftP = heapP + heap_index;
					if (top_clones_endP <= leftP) {
						heapP->weight = 0xFFFFFFFF;
						break;
					}
					rightP = leftP + 1;
					if (top_clones_endP <= rightP || leftP->weight <= rightP->weight) {
						*heapP       = *leftP;
						heapP        = leftP;
						heap_index <<= 1;
						continue;
					}
					*heapP     = *rightP;
					heapP      = rightP;
					heap_index = (heap_index << 1) + 1;
			}	}
			sprintf(g_filename,"%s/top.html", g_html_dirP);
			topF = open_file(g_filename);
			fprintf(topF, HTML_HEADER "<title>Top %d largest clones</title>\n<body>\n<h3>Top %d largest clones</h3>\n<p>\n<table><tr>\n", cnt, cnt);
			for (i = 0; i < g_top_cols; ++i) {
				fprintf(topF, "<th><font color=red>Weight</font></th><th>Clone</th>");
			}
			for (i = g_top_cols; --atP >= topP; ++i) {
				if (i == g_top_cols) {
					fprintf(topF, "</tr>\n<tr>");
					i = 0;
				}
				fprintf(topF, "<td align=right>%u</td><td><a href=%u.html>%u.html</td>", atP->weight, atP->html, atP->html);
			}
			fprintf(topF, "</tr>\n</table>\n<P>\nEnd of report\n</body>\n</html>\n");
			fclose(topF);
	}	}
}

static int
permitted_clone(void)
{
	if ( (g_lth1 < g_min_match_sequence) || (g_lth2 < g_min_match_sequence)) {
		// Generally the clone is too short on one side or the other to be deemed worth reporting
		// Consider a clone long enough if it includes all of a function
		if ((start_code1P != function_code1P || best_end1P != function_end1P || g_lth1 < g_min_function_sequence) &&
			(start_code2P != function_code2P || best_end2P != function_end2P || g_lth2 < g_min_function_sequence)
		   ) {
			return(0);
	}	}
	return(1);
}

static int
same_branch(codeT *code1P, codeT *code2P)
{
	codeT	*to1P, *to2P, *matchP;

	to1P = code1P->m_jump.P;
	to2P = code2P->m_jump.P;
	if (to1P < code1P) {
		if (to2P >= code2P) {
			goto mismatch;
		}
		if (to1P < start_code1P) {
			goto mismatch;
		}
		if (to2P < start_code2P) {
			goto mismatch;
		}
		if ((matchP = to1P->m_matchP)) {
			if (to2P > matchP) {
				goto mismatch;
			}
			if (to2P->m_matchP) {
				if (to2P == matchP) {
					goto same;
				}
				goto mismatch;
			}
		} else if ((matchP = to2P->m_matchP)) {
			if (to1P > matchP) {
				goto mismatch;
			}
			if (to2P == start_code2P) {
				goto mismatch;
		}	}

		while (!((--to1P)->m_matchP));
		assert(start_code1P <= to1P);
		while (!((--to2P)->m_matchP));
		assert(start_code2P <= to2P);

		if (to1P->m_matchP != to2P) {
			goto mismatch;
		}
	} else if (to1P > code1P) {
		if (to2P <= code2P) {
			goto mismatch;
		}
		if (end_code1P <= to1P) {
			goto mismatch;
		}
		if (end_code2P <= to2P) {
			goto mismatch;
		}
	} else if (to2P != code2P) {
		goto mismatch;
	}
same:
	return(1);
mismatch:
	return(0);
}

static int
same_register(registerT *register1P, registerT *register2P)
{
	variableT	*variable1P, *variable2P;
	stringT		*name1P, *name2P;

	if ((variable1P = register1P->u.variableP) && (variable2P = register2P->u.variableP)) {
		if ((name1P = variable1P->nameP) && (name2P = variable2P->nameP)) {
			return(variable1P->nameP->P == variable2P->nameP->P);
	}	}
	return (register1P->number == register2P->number);
}

static int
same_string(stringT *string1P, stringT *string2P)
{
	return (string1P->P == string2P->P);
}

static int
same_type(stringT *string1P, stringT *string2P)
{
	return (string1P->P == string2P->P);
}

static int
same_field(fieldT *field1P, fieldT *field2P)
{
	return(field1P->nameP == field2P->nameP);
}

static int
same_method(methodT *method1P, methodT *method2P)
{
	return(method1P->nameP == method2P->nameP);
}

static int
same_code(codeT *code1P, codeT *code2P, int *weightP) 
{
	pcodeE	pcode1;

	pcode1 = code1P->m_pcode;
	if (pcode1 != code2P->m_pcode) {
		goto mismatch;
	}

	switch (pcode1) {
	case NOP:				// 0x00: 10x ->  00|op  -> op
	case RETURN_VOID:		// 0x0e:
		break;

	case MOVE_RESULT:		// 0x0a: 11x ->  AA|op  -> op vAA 
	case MOVE_RESULT_WIDE:	// 0x0b
	case MOVE_RESULT_OBJECT:// 0x0c
	case MOVE_EXCEPTION:	// 0x0d
	case RETURN:			// 0x0f
	case RETURN_WIDE:		// 0x10
	case RETURN_OBJECT:		// 0x11
	case MONITOR_ENTER:		// 0x1d
	case MONITOR_EXIT:		// 0x1e
	case THROW:				// 0x27
		if (!same_register(&(code1P->info.register1P->reg1), &(code2P->info.register1P->reg1))) {
			goto mismatch;
		}
		break;
	case ARRAY_LENGTH:		// 0x21: 12x -> B|A|op -> op vA, vB 
	case NEG_INT:			// 0x7b:
	case NOT_INT:			// 0x7c:
	case NEG_LONG:			// 0x7d:
	case NOT_LONG:			// 0x7e:
	case NEG_FLOAT:			// 0x7f:
	case NEG_DOUBLE:		// 0x80:
	case INT_TO_LONG:		// 0x81:
	case INT_TO_FLOAT:		// 0x82:
	case INT_TO_DOUBLE:		// 0x83:
	case LONG_TO_INT:		// 0x84:
	case LONG_TO_FLOAT:		// 0x85:
	case LONG_TO_DOUBLE:	// 0x86:
	case FLOAT_TO_INT:		// 0x87:
	case FLOAT_TO_LONG:		// 0x88:
	case FLOAT_TO_DOUBLE:	// 0x89:
	case DOUBLE_TO_INT:		// 0x8a:
	case DOUBLE_TO_LONG:	// 0x8b:
	case DOUBLE_TO_FLOAT:	// 0x8c:
	case INT_TO_BYTE:		// 0x8d:
	case INT_TO_CHAR:		// 0x8e:
	case INT_TO_SHORT:		// 0x8f:
	case ADD_INT2:			// 0xb0:
	case SUB_INT2:			// 0xb1:
	case MUL_INT2:			// 0xb2:
	case DIV_INT2:			// 0xb3:
	case REM_INT2:			// 0xb4:
	case AND_INT2:			// 0xb5:
	case OR_INT2:			// 0xb6:
	case XOR_INT2:			// 0xb7:
	case SHIFT_LEFT_INT2:	// 0xb8:
	case SHIFT_RIGHT_INT2:	// 0xb9:
	case USHIFT_RIGHT_INT2:	// 0xba:
	case ADD_LONG2:			// 0xbb:
	case SUB_LONG2:			// 0xbc:
	case MUL_LONG2:			// 0xbd:
	case DIV_LONG2:			// 0xbe:
	case REM_LONG2:			// 0xbf:
	case AND_LONG2:			// 0xc0:
	case OR_LONG2:			// 0xc1:
	case XOR_LONG2:			// 0xc2:
	case SHIFT_LEFT_LONG2:	// 0xc3:
	case SHIFT_RIGHT_LONG2:	// 0xc4:
	case USHIFT_RIGHT_LONG2:// 0xc5:
	case ADD_FLOAT2:		// 0xc6:
	case SUB_FLOAT2:		// 0xc7:
	case MUL_FLOAT2:		// 0xc8:
	case DIV_FLOAT2:		// 0xc9:
	case REM_FLOAT2:		// 0xca:
	case ADD_DOUBLE2:		// 0xcb:
	case SUB_DOUBLE2:		// 0xcc:
	case MUL_DOUBLE2:		// 0xcd:
	case DIV_DOUBLE2:		// 0xce:
	case REM_DOUBLE2:		// 0xcf:

	case MOVE_4_4:			// 0x01: 12x -> B|A|op -> op vA, vB
	case MOVE_WIDE_4_4:		// 0x04
	case MOVE_OBJECT_4_4:	// 0x07
	case MOVE_8_16:			// 0x02: 22x -> AA|op BBBB -> op vAA, vBBBB
	case MOVE_WIDE_8_16:	// 0x05:
	case MOVE_OBJECT_8_16:	// 0x08:
	case MOVE_16_16:		// 0x03: 32x -> |op AAAA BBBB -> op vAAAA, vBBBB
	case MOVE_WIDE_16_16:	// 0x06:
	case MOVE_OBJECT_16_16:	// 0x09:
	{
		info_register2T	*registers1P, *registers2P;

		registers1P = code1P->info.register2P;
		registers2P = code2P->info.register2P;
		if (!same_register(&(registers1P->reg1), &(registers2P->reg1))) {
			goto mismatch;
		}
		if (!same_register(&(registers1P->reg2), &(registers2P->reg2))) {
			goto mismatch;
		}
		break;
	}
	case CMPL_FLOAT:		// 0x2d: 23x -> AA|op CC|BB -> op vAA, vBB, vCC
	case CMPG_FLOAT:		// 0x2e:
	case CMPL_DOUBLE:		// 0x2f:
	case CMPG_DOUBLE:		// 0x30:
	case CMP_LONG:			// 0x31:

	case AGET:				// 0x44:
	case AGET_WIDE:			// 0x45:
	case AGET_OBJECT:		// 0x46:
	case AGET_BOOLEAN:		// 0x47:
	case AGET_BYTE:			// 0x48:
	case AGET_CHAR:			// 0x49:
	case AGET_SHORT:		// 0x4a:
	case APUT:				// 0x4b:
	case APUT_WIDE:			// 0x4c:
	case APUT_OBJECT:		// 0x4d:
	case APUT_BOOLEAN:		// 0x4e:
	case APUT_BYTE:			// 0x4f:
	case APUT_CHAR:			// 0x50:
	case APUT_SHORT:		// 0x51:

	case ADD_INT:			// 0x90:
	case SUB_INT:			// 0x91:
	case MUL_INT:			// 0x92:
	case DIV_INT:			// 0x93:
	case REM_INT:			// 0x94:
	case AND_INT:			// 0x95:
	case OR_INT:			// 0x96:
	case XOR_INT:			// 0x97:
	case SHIFT_LEFT_INT:	// 0x98:
	case SHIFT_RIGHT_INT:	// 0x99:
	case USHIFT_RIGHT_INT:	// 0x9a:
	case ADD_LONG:			// 0x9b:
	case SUB_LONG:			// 0x9c:
	case MUL_LONG:			// 0x9d:
	case DIV_LONG:			// 0x9e:
	case REM_LONG:			// 0x9f:
	case AND_LONG:			// 0xa0:
	case OR_LONG:			// 0xa1:
	case XOR_LONG:			// 0xa2:
	case SHIFT_LEFT_LONG:	// 0xa3:
	case SHIFT_RIGHT_LONG:	// 0xa4:
	case USHIFT_RIGHT_LONG:	// 0xa5:
	case ADD_FLOAT:			// 0xa6:
	case SUB_FLOAT:			// 0xa7:
	case MUL_FLOAT:			// 0xa8:
	case DIV_FLOAT:			// 0xa9:
	case REM_FLOAT:			// 0xaa:
	case ADD_DOUBLE:		// 0xab:
	case SUB_DOUBLE:		// 0xac:
	case MUL_DOUBLE:		// 0xad:
	case DIV_DOUBLE:		// 0xae:
	case REM_DOUBLE:		// 0xaf:
	{
		info_register3T	*registers1P, *registers2P;

		registers1P = code1P->info.register3P;
		registers2P = code2P->info.register3P;
		if (!same_register(&(registers1P->reg1), &(registers2P->reg1))) {
			goto mismatch;
		}
		if (!same_register(&(registers1P->reg2), &(registers2P->reg2))) {
			goto mismatch;
		}
		if (!same_register(&(registers1P->reg3), &(registers2P->reg3))) {
			goto mismatch;
		}
		break;
	}
	case CONST_4_4:			// 0x12: 11n -> B|A|op -> op vA, #+B 
	case CONST_8_16:		// 0x13: 21s -> AA|op BBBB -> op vAA, #+BBBB
	case CONST_8_32:		// 0x14: 31i -> AA|op BBBBlo BBBBhi -> op vAA, #+BBBBBBBB
	case CONST_HIGH:		// 0x15: 21h -> AA|op BBBB -> op vAA, #+BBBB0000
	case CONST_WIDE_HIGH:	// 0x19: 21h -> AA|op BBBB -> op vAA, #+BBBB000000000000 
	{
		info_valueT	*registers1P, *registers2P;

		registers1P = code1P->info.valueP;
		registers2P = code2P->info.valueP;		
		if (registers1P->value != registers2P->value) {
			goto mismatch;
		}
		if (!same_register(&(registers1P->reg1), &(registers2P->reg1))) {
			goto mismatch;
		}
		break;
	}
	case CONST_WIDE_8_16:	// 0x16: 21s -> const-wide/16 vAA, #+BBBB
	case CONST_WIDE_8_32:	// 0x17:
	case CONST_WIDE_8_64:	// 0x18: 51l -> AA|op BBBBlo BBBB BBBB BBBBhi -> op vAA, #+BBBBBBBBBBBBBBBB
	{
		info_wideT	*registers1P, *registers2P;

		registers1P = code1P->info.wideP;
		registers2P = code2P->info.wideP;	
		if (registers1P->value != registers2P->value) {
			goto mismatch;
		}
		if (!same_register(&(registers1P->reg1), &(registers2P->reg1))) {
			goto mismatch;
		}
		break;
	}
	case ADD_INT_LIT8:		// 0xd8: 22b -> AA|op CC|BB -> op vAA, vBB, #+CC
	case REVERSE_SUB_INT_LIT8: // 0xd9:
	case MUL_INT_LIT8:		// 0xda:
	case DIV_INT_LIT8:		// 0xdb:
	case REM_INT_LIT8:		// 0xdc:
	case AND_INT_LIT8:		// 0xdd:
	case OR_INT_LIT8:		// 0xde:
	case XOR_INT_LIT8:		// 0xdf:
	case SHIFT_LEFT_INT_LIT8://0xe0:
	case SHIFT_RIGHT_INT_LIT8://0xe1:
	case USHIFT_RIGHT_INT_LIT8://0xe2:

	case ADD_INT_LIT16:		// 0xd0: 22s -> B|A|op CCCC -> op vA, vB, #+CCCC
	case REVERSE_SUB_INT_LIT16:	// 0xd1:
	case MUL_INT_LIT16:		// 0xd2:
	case DIV_INT_LIT16:		// 0xd3:
	case REM_INT_LIT16:		// 0xd4:
	case AND_INT_LIT16:		// 0xd5:
	case OR_INT_LIT16:		// 0xd6:
	case XOR_INT_LIT16:		// 0xd7:
	{
		info_value2T	*registers1P, *registers2P;

		registers1P = code1P->info.value2P;
		registers2P = code2P->info.value2P;	
		if (registers1P->value != registers2P->value) {
			goto mismatch;
		}
		if (!same_register(&(registers1P->reg1), &(registers2P->reg1))) {
			goto mismatch;
		}
		if (!same_register(&(registers1P->reg2), &(registers2P->reg2))) {
			goto mismatch;
		}
		break;
	}
	case CONST_STRING_8_16: // 0x1a: 21c -> AA|op BBBB -> op vAA, string@BBBB
	case CONST_STRING_8_32:	// 0x1b: 31c -> AA|op BBBBlo BBBBhi -> op vAA, string@BBBBBBBB
	{
		info_register2T	*registers1P, *registers2P;

		registers1P = code1P->info.register2P;
		registers2P = code2P->info.register2P;
		if (!same_register(&(registers1P->reg1), &(registers2P->reg1))) {
			goto mismatch;
		}
		if (!same_string(registers1P->reg2.u.stringP, registers2P->reg2.u.stringP)) {
			goto mismatch;
		}
		break;
	}
	case CONST_CLASS:		// 0x1c: 21c -> AA|op BBBB -> op vAA, type@BBBB
	case CHECK_CAST:		// 0x1f:
	case NEW_INSTANCE:		// 0x22:
	{
		info_register2T	*registers1P, *registers2P;

		registers1P = code1P->info.register2P;
		registers2P = code2P->info.register2P;

		if (!same_register(&(registers1P->reg1), &(registers2P->reg1))) {
			goto mismatch;
		}
		if (!same_type(registers1P->reg2.u.typeP, registers2P->reg2.u.typeP)) {
			goto mismatch;
		}
		break;
	}
	case STATIC_GET:		// 0x60: 21c -> AA|op BBBB -> op vAA, field@BBBB
	case STATIC_GET_WIDE:	// 0x61:
	case STATIC_GET_OBJECT:	// 0x62:
	case STATIC_GET_BOOLEAN:// 0x63:
	case STATIC_GET_BYTE:	// 0x64:
	case STATIC_GET_CHAR:	// 0x65:
	case STATIC_GET_SHORT:	// 0x66:
	case STATIC_PUT:		// 0x67:
	case STATIC_PUT_WIDE:	// 0x68:
	case STATIC_PUT_OBJECT:	// 0x69:
	case STATIC_PUT_BOOLEAN:// 0x6a:
	case STATIC_PUT_BYTE:	// 0x6b:
	case STATIC_PUT_CHAR:	// 0x6c:
	case STATIC_PUT_SHORT:	// 0x6d:
	{
		info_register2T	*registers1P, *registers2P;

		registers1P = code1P->info.register2P;
		registers2P = code2P->info.register2P;

		if (!same_register(&(registers1P->reg1), &(registers2P->reg1))) {
			goto mismatch;
		}
		if (!same_field(registers1P->reg2.u.fieldP, registers2P->reg2.u.fieldP)) {
			goto mismatch;
		}
		break;
	}
	case INSTANCE_OF:		// 0x20: 22c -> B|A|op CCCC -> op vA, vB, type@CCCC
	case NEW_ARRAY:			// 0x23:
	{
		info_register3T	*registers1P, *registers2P;

		registers1P = code1P->info.register3P;
		registers2P = code2P->info.register3P;
		if (!same_register(&(registers1P->reg1), &(registers2P->reg1))) {
			goto mismatch;
		}
		if (!same_register(&(registers1P->reg2), &(registers2P->reg2))) {
			goto mismatch;
		}
		if (!same_type(registers1P->reg3.u.typeP, registers2P->reg3.u.typeP)) {
			goto mismatch;
		}
		break;
	}
	case IGET:				// 0x52: 22c -> B|A|op CCCC -> op vA, vB, field@CCCC
	case IGET_WIDE:			// 0x53:
	case IGET_OBJECT:		// 0x54:
	case IGET_BOOLEAN:		// 0x55:
	case IGET_BYTE:			// 0x56:
	case IGET_CHAR:			// 0x57:
	case IGET_SHORT:		// 0x58:
	case IPUT:				// 0x59:
	case IPUT_WIDE:			// 0x5a:
	case IPUT_OBJECT:		// 0x5b:
	case IPUT_BOOLEAN:		// 0x5c:
	case IPUT_BYTE:			// 0x5d:
	case IPUT_CHAR:			// 0x5e:
	case IPUT_SHORT:		// 0x5f:
	{
		info_register3T	*registers1P, *registers2P;

		registers1P = code1P->info.register3P;
		registers2P = code2P->info.register3P;

		if (!same_register(&(registers1P->reg1), &(registers2P->reg1))) {
			goto mismatch;
		}
		if (!same_register(&(registers1P->reg2), &(registers2P->reg2))) {
			goto mismatch;
		}
		if (!same_field(registers1P->reg3.u.fieldP, registers2P->reg3.u.fieldP)) {
			goto mismatch;
		}
		break;
	}
	case GOTO_8:			// 0x28: 10t ->  AA|op -> op +AA
	case GOTO_16:			// 0x29: 20t -> |op AAAA -> op +AAAA
	case GOTO_32:			// 0x2a:		// 30t -> |op AAAAlo AAAAhi -> op +AAAAAAAA
		if (!same_branch(code1P, code2P)) {
			goto mismatch;
		}
		break;

	case IF_EQ_ZERO:		// 0x38: 21t -> AA|op BBBB -> op vAA, +BBBB
	case IF_NE_ZERO:		// 0x39:
	case IF_LT_ZERO:		// 0x3a:
	case IF_GE_ZERO:		// 0x3b:
	case IF_GT_ZERO:		// 0x3c:
	case IF_LE_ZERO:		// 0x3d:
	{
		info_register1T	*registers1P, *registers2P;

		registers1P = code1P->info.register1P;
		registers2P = code2P->info.register1P;
		if (!same_register(&(registers1P->reg1), &(registers2P->reg1))) {
			goto mismatch;
		}
		if (!same_branch(code1P, code2P)) {
			goto mismatch;
		}
		break;
	}
	case IF_EQ:				// 0x32: 22t -> B|A|op CCCC -> op vA, vB, +CCCC
	case IF_NE:				// 0x33:
	case IF_LT:				// 0x34:
	case IF_GE:				// 0x35:
	case IF_GT:				// 0x36:
	case IF_LE:				// 0x37:
	{
		info_register2T	*registers1P, *registers2P;

		registers1P = code1P->info.register2P;
		registers2P = code2P->info.register2P;
		if (!same_register(&(registers1P->reg1), &(registers2P->reg1))) {
			goto mismatch;
		}
		if (!same_register(&(registers1P->reg2), &(registers2P->reg2))) {
			goto mismatch;
		}
		if (!same_branch(code1P, code2P)) {
			goto mismatch;
		}
		break;
	}

	case FILL_ARRAY_DATA:	// 0x26: 31t -> AA|op BBBBlo BBBBhi -> op vAA, +BBBBBBBB fill-array-data vAA, +BBBBBBBB (with supplemental data as specified below in "fill-array-data Format")
	{
		info_fillT	*registers1P, *registers2P;
		u4T width;
		u4T entries;

		registers1P = code1P->info.fillP;
		registers2P = code2P->info.fillP;

		if (!same_register(&(registers1P->reg1), &(registers2P->reg1))) {
			goto mismatch;
		}
		width = registers1P->width;
		if (width   != registers2P->width) {
			goto mismatch;
		}
		entries = registers1P->entries;
		if (entries != registers2P->entries) {
			goto mismatch;
		}
		if (memcmp(registers1P->data, registers2P->data, width * entries)) {
			goto mismatch;
		}
		break;
	}
	case PACKED_SWITCH:		// 0x2b: 31t -> AA|op BBBBlo BBBBhi -> op vAA, +BBBBBBBB (with supplemental data as specified below in "packed-switch Format")
	case SPARSE_SWITCH:		// 0x2c: 31t -> AA|op BBBBlo BBBBhi -> op vAA, +BBBBBBBB  (with supplemental data as specified below in "sparse-switch Format")
	{	
		info_valueT	*registers1P, *registers2P;

		registers1P = code1P->info.valueP;
		registers2P = code2P->info.valueP;
		if (registers1P->value != registers2P->value) {
			goto mismatch;
		}		
		if (!same_register(&(registers1P->reg1), &(registers2P->reg1))) {
			goto mismatch;
		}
		if (!same_branch(code1P, code2P)) {
			goto mismatch;
		}
		break;
	}
	case FILL_ARRAY:		// 0x24: 35c -> B|A|op CCCC G|F|E|D -> [B=5] op {vD, vE, vF, vG, vA}, type@CCCC
	case FILLED_NEW_ARRAY:	// 0x25: 3rc -> AA|op BBBB CCCC -> op {vCCCC .. vNNNN}, type@BBBB
	{
		info_registersT	*registers1P, *registers2P;
		u4T				a1, a2, i;

		registers1P = code1P->info.registersP;
		registers2P = code2P->info.registersP;
		a1 = registers1P->size;
		a2 = registers2P->size;
		if (a1 != a2) {
			// Not same number of arguments
			goto mismatch;
		}
		if (!same_type(registers1P->last.u.typeP, registers2P->last.u.typeP)) {
			goto mismatch;
		}
		for (i = 0; i < a1; ++i) {
			if (!same_register(registers1P->reg + i, registers2P->reg + i)) {
				goto mismatch;
		}	}
		// Can't know what the value of the registers are
		break;
	}
	case INVOKE_VIRTUAL_RANGE: // 0x74: 3rc -> AA|op BBBB CCCC -> op {vCCCC .. vNNNN}, meth@BBBB
	case INVOKE_SUPER_RANGE://0x75:
	case INVOKE_DIRECT_RANGE://0x76:
	case INVOKE_STATIC_RANGE://0x77:
	case INVOKE_INTERFACE_RANGE://0x78:

	case INVOKE_VIRTUAL:	// 0x6e: [B=5] op {vD, vE, vF, vG, vA}, meth@CCCC
	case INVOKE_SUPER:		// 0x6f: [B=4] op {vD, vE, vF, vG}, meth@CCCC
	case INVOKE_DIRECT:		// 0x70: [B=3] op {vD, vE, vF}, meth@CCCC
	case INVOKE_STATIC:		// 0x71: [B=2] op {vD, vE}, meth@CCCC
	case INVOKE_INTERFACE:	// 0x72: [B=1] op {vD}, meth@CCCC
					//				 [B=0] op {}, meth@CCCC
	{
		info_registersT	*registers1P, *registers2P;
		u4T				a1, a2, i;

		registers1P = code1P->info.registersP;
		registers2P = code2P->info.registersP;
		a1 = registers1P->size;
		a2 = registers2P->size;
		if (a1 != a2) {
			// Not same number of arguments
			goto mismatch;
		}
		if (!same_method(registers1P->last.u.methodP, registers2P->last.u.methodP)) {
			goto mismatch;
		}
		for (i = 0; i < a1; ++i) {
			if (!same_register(registers1P->reg + i, registers2P->reg + i)) {
				goto mismatch;
		}	}
		// Can't know what the value of the registers are
		break;
	}
	case TRY_START:
	{
		info_tryT	*registers1P, *registers2P;

		registers1P = code1P->info.tryP;
		registers2P = code2P->info.tryP;
		if (!registers1P->classtype.u.typeP) {
			if (registers2P->classtype.u.typeP) {
				goto mismatch;
			}
		} else {
			if (!registers2P->classtype.u.typeP) {
				goto mismatch;
			}
			if (!same_type(registers1P->classtype.u.typeP, registers2P->classtype.u.typeP)) {
				goto mismatch;
		}	}
		if (!same_branch(code1P, code2P)) {
			goto mismatch;
		}
		break;
	}
	case TRY_END:
	{
		info_tryT	*registers1P, *registers2P;

		registers1P = code1P->info.tryP;
		registers2P = code2P->info.tryP;
		if (!registers1P->classtype.u.typeP) {
			if (registers2P->classtype.u.typeP) {
				goto mismatch;
			}
		} else {
			if (!registers2P->classtype.u.typeP) {
				goto mismatch;
			}
			if (!same_type(registers1P->classtype.u.typeP, registers2P->classtype.u.typeP)) {
				goto mismatch;
		}	}		break;
	}
	default:
		assert(0);
		goto mismatch;
	}
	*weightP = g_same_cost;
	return(1);
mismatch:
	*weightP = g_mismatch_cost;
	return(0);
}

// Hill Climbing Logic

typedef struct dllS {
	struct dllS		*prevP;
	struct dllS 	*nextP;
} dllT;

typedef unsigned short crossT;

typedef	struct edgeS {
	dllT			dll;			// Must be first element in struct
	codeT			*code1P;		// Left  side of match
	codeT			*code2P;		// Right side of match
	int				matched;		// This is a matching edge
	int				blocks;			// Number of edges that block this edge
	struct edgeS	*next_changeP;
	int				blocks1;

	int				crosses;
	crossT			*crossesP;
} edgeT;

typedef struct {
	edgeT			*edgeP;
	int				matched;
} reverseT;

typedef struct {
	edgeT			*edgeP;
	int				blocks;
} unblockedT;

static int
hill_climb(void)
{
	// Array of edges
	static	edgeT	*edgesP         = 0;
	static	edgeT	*edges_maxP     = 0;
	static	unsigned int		max_edges       = 0;
	static	edgeT	*end_edgesP     = 0;

	// Array of crossings of edges with edges
	static	crossT	*crossesP       = 0;
	static	unsigned int	max_crosses     = 0;

#define Backup(X) \
	if (X->next_changeP == (edgeT *) -1) { \
		X->blocks1      = X->blocks; \
		X->next_changeP = changesP; \
		changesP        = X; \
	}

#define Increment(X) \
	Backup(X); \
	++(X->blocks1);

#define Decrement(X) \
	Backup(X); \
	--(X->blocks1);

	// Array of edges whose matched/unmatched status is to be reversed
	static	reverseT		*reversesP     = 0;
	static	reverseT		*reverses_maxP = 0;
	static	unsigned int	max_reverses   = 0;
	static	reverseT		*end_reversesP = 0;

#define Reverse(X) \
	if (reverses_maxP <= end_reversesP) { \
		at = max_reverses; \
		if (!max_reverses) { \
			max_reverses   = 4096; \
		} else { \
			max_reverses <<= 1; \
		} \
		reversesP     = (reverseT *) Xrealloc(reversesP, max_reverses * sizeof(reverseT)); \
		if (!reversesP) { \
            fprintf(stderr, "Hillclimb: Reversals outofmemory\n"); \
			reverses_maxP = 0; \
            max_reverses  = 0; \
            goto abort_reversals; \
        } \
		reverses_maxP = reversesP + max_reverses; \
		end_reversesP = reversesP + at; \
	} \
	end_reversesP->edgeP   = X; \
	end_reversesP->matched = X->matched; \
	++end_reversesP; 

	static	unblockedT	*unblockedP = 0;
	static	unblockedT	*unblocked_maxP = 0;
	static	unsigned int	 max_unblocked  = 0;
	static	unblockedT	*end_unblockedP = 0;

	// Array of dll's for each block which we wish to handle
	static	dllT 	*dllsP     = 0;
	static	dllT	*end_dllsP = 0;

	// Linked list of edges whose blocks value changes
	static	edgeT	*changesP  = 0;

	static	int		iterations;

	static codeT	*code1aP, *code2aP, *code1bP, *code2bP;
	static crossT	*cross1P, *cross2P, *cross3P;
	static crossT	*end_cross1P, *end_cross2P, *end_cross3P;
	static edgeT	*edge1P, *next_edge1P;
	static edgeT	*edge2P, *edge3P, *edge4P;
	static unblockedT *unblocked1P, *unblocked_bestP;

	static int		at, phase, max_crossings, change, old_ret, ret;
	static int		weight;
	static int		matched;
	static reverseT *reverseP;
	static int		blocks;
	static dllT		*headP, *cycleP;
	static unsigned int crosses;

	static time_t	hillclimb_started;
	static time_t	hillclimb_terminate;
	static methodT	*method1P, *method2P;

#ifdef TEST_HILLCLIMB
	static int		original_matches, final_matches;
#endif

	hillclimb_started   = time(0);
	if (g_hillclimb_timeout) {
		hillclimb_terminate = hillclimb_started + g_hillclimb_timeout;
	} else {
		hillclimb_terminate = 0;
    }
	g_hillclimb_elapse -= hillclimb_started;

#ifdef DEBUG_HILLCLIMB
	fprintf(stderr, "Hill Climb %p %d %p %d\n", 
		start_code1P, 
		best_end1P - start_code1P,
		start_code2P, 
		best_end2P - start_code2P);
#endif
#ifdef TEST_HILLCLIMB
	original_matches = 0;
	for (code1aP = start_code1P; code1aP <= best_end1P; ++code1aP) {
		if (code1aP->m_matchP) {
			++original_matches;
	}	}
#endif

	if (!dllsP) {
		dllsP     = (dllT *) Xmalloc(g_hillclimb * sizeof(dllT));
		if (!dllsP) {
			fprintf(stderr, "Hillclimb: dlls outofmemory\n");
			return(0);
		}
		end_dllsP = dllsP + g_hillclimb;
	}

	// Find all possible edges

	end_edgesP       = edgesP;

	// Compute all the edges
	// These may change for jumps as matchings change

	// Can't change first line
	change    = 0;
	for (code1aP = start_code1P + 1; code1aP <= best_end1P; ++code1aP) {
		for (code2aP = start_code2P + 1; code2aP <= best_end2P; ++code2aP) {
			// Add the edge if appropriate
			if (code1aP->m_matchP == code2aP) {
				matched = 1;
			} else {
				if (!same_code(code1aP, code2aP, &weight)) {
					// Not appropriate
					continue;
				}
				matched = 0;
				change  = 1;
			}
			if (edges_maxP <= end_edgesP) {
				at = max_edges;
				if (!max_edges) {
					max_edges   = (1<<11);
				} else {
					max_edges <<= 1;
				}
				edgesP     = (edgeT *) Xrealloc(edgesP, max_edges * sizeof(edgeT));
				if (!edgesP) {
					max_edges = 0;
					goto too_many_edges;
				}
				edges_maxP = edgesP + max_edges;
				end_edgesP = edgesP + at;
			}
			memset(end_edgesP, 0, sizeof(edgeT));
			end_edgesP->code1P       = code1aP;
			end_edgesP->code2P       = code2aP;
			end_edgesP->matched      = matched;
			end_edgesP->next_changeP = (edgeT *) -1;
			++end_edgesP;
	}	}

	if (!change) {
		// No unmatched edges
		ret = 0;
		goto done;
	}

	if ((end_edgesP - edgesP) > 0xFFFF) {
too_many_edges:
		method1P = function1P->m_methodP;
		method2P = function2P->m_methodP;
		// Can't handle the edge offsets in a short
		fprintf(stderr,
			"* Hill-climbing aborted: too many edges\n* ");
		dump_wchar(method1P->classP->P, stderr);
		fputc(':', stderr);
		dump_wchar(method1P->nameP->P, stderr);
		fputs(" .v.\n* ", stderr);
		dump_wchar(method2P->classP->P, stderr);
		fputc(':', stderr);
		dump_wchar(method2P->nameP->P, stderr);
		fputc('\n', stderr);
		fprintf(stderr, "* Clone lths={%ld,%ld}, edges=%ld\n",
			(long) (best_end1P - start_code1P),
			(long) (best_end2P - start_code2P),
			(long) (end_edgesP - edgesP));
		ret = 0;
		goto done;
	}

	// Add edge crossings

	crosses = 0;
	for (phase = 0; ; ++phase) {
		for (edge1P = edgesP; edge1P < end_edgesP; ++edge1P) {
			code1aP = edge1P->code1P;
			code2aP = edge1P->code2P;
			// Treat an edge point match as a cross

			for (edge2P = edge1P; ++edge2P < end_edgesP; ) {
				code1bP = edge2P->code1P;
				code2bP = edge2P->code2P;
				if ((code1aP >= code1bP && code2aP <= code2bP) ||
					(code1aP <= code1bP && code2aP >= code2bP)) {
					if (!phase) {
						++(edge1P->crosses);
						++(edge2P->crosses);
						crosses += 2;
					} else {
						--(edge1P->crossesP);
						*edge1P->crossesP = (crossT) (edge2P - edgesP);
						--(edge2P->crossesP);
						*edge2P->crossesP = (crossT) (edge1P - edgesP);
						if (edge1P->matched) {
							assert(!edge2P->matched);
							++(edge2P->blocks);
						} else if (edge2P->matched) {
							++(edge1P->blocks);
		}	}	}	}	}
		if (phase) {
			break;
		}
		if (!crosses) {
			ret = 0;
			goto done;
		}
		if (crosses > 0x40000000) {
too_many_crosses:
			method1P = function1P->m_methodP;
			method2P = function2P->m_methodP;

			fprintf(stderr, "* Hill-climbing aborted: too many crosses\n* ");
			dump_wchar(method1P->classP->P, stderr);
			fputc(':', stderr);
			dump_wchar(method1P->nameP->P, stderr);
			fputs(" .v.\n* ", stderr);
			dump_wchar(method2P->classP->P, stderr);
			fputc(':', stderr);
			dump_wchar(method2P->nameP->P, stderr);
			fputc('\n', stderr);
			fprintf(stderr, "* Clone lths={%ld,%ld}, edges=%ld crosses=%u\n",
				(long) (best_end1P - start_code1P),
				(long) (best_end2P - start_code2P),
				(long) (end_edgesP - edgesP),
				crosses);
			ret = 0;
			goto done;
		}

		if (max_crosses < crosses) {
			if (!max_crosses) {
				max_crosses = (1<<18);	// Must be multiple of 2
			} else {
				Xfree(crossesP);
			}
			while (max_crosses < crosses) {
				max_crosses <<= 1;
			}
			crossesP = (crossT *) Xmalloc(max_crosses * sizeof(unsigned short));
			if (!crossesP) {
				max_crosses = 0;
				goto too_many_crosses;
			}
		}
		end_cross1P = crossesP + crosses;
		for (edge1P = end_edgesP; edgesP <= --edge1P; ) {
			edge1P->crossesP = end_cross1P;
			end_cross1P     -= edge1P->crosses;
		}
		assert(end_cross1P == crossesP);
	}

	for (headP = dllsP; headP < end_dllsP; ++headP) {
		headP->nextP = headP;
		headP->prevP = headP;
	}

	for (edge2P = edgesP; edge2P < end_edgesP; ++edge2P) {
#ifdef TEST_HILLCLIMB
		assert(edge2P->blocks || edge2P->matched);
#endif
		blocks = edge2P->blocks;
		if (blocks > g_max_blocks_seen) {
			g_max_blocks_seen = blocks;
		}
		if (blocks && blocks <= g_hillclimb) {
			// Add to the start of the cycle
			--blocks;
			headP                    = dllsP + blocks;
			edge2P->dll.prevP        = headP;
			edge2P->dll.nextP        = headP->nextP;
			headP->nextP             = (dllT *) edge2P;
			edge2P->dll.nextP->prevP = (dllT *) edge2P;
		} else {
			edge2P->dll.nextP        = 0;
			edge2P->dll.prevP        = 0;
	}	}

#ifdef DEBUG_HILLCLIMB
	fprintf(stderr, "%d edges %d edge crossings %d code1=%p %d code2=%p\n", 
				end_edgesP - edgesP,
				crosses,
				best_end1P - start_code1P,
				start_code1P,
				best_end2P - start_code2P,
				start_code2P);

	for (edge1P = edgesP; edge1P < end_edgesP; ++edge1P) {
		fprintf(stderr, "%d] %d %d<->%d %s<=>%s ", 
					edge1P - edgesP,
					edge1P->blocks,
					edge1P->code1P - start_code1P,
					edge1P->code2P - start_code2P,
					edge1P->code1P->m_codeP,
					edge1P->code2P->m_codeP);
		if (edge1P->matched) {
			fprintf(stderr, " match");
		}
		cross1P = edge1P->crossesP;
		for (end_cross1P = cross1P + edge1P->crosses; cross1P < end_cross1P; ++cross1P) {
			fprintf(stderr, " %hu", *cross1P);
		}
		fprintf(stderr, "\n");
	}
#endif

	old_ret = ret  = 0;
	iterations     = 0;
	for (max_crossings = 0; ; ++max_crossings) {
		if (max_crossings >= g_hillclimb) {
			if (ret == old_ret || iterations >= g_iterations) {
				if (iterations > g_max_iterations_seen) {
					g_max_iterations_seen = iterations;
				}
				break;
			}
			++iterations;
			old_ret        = ret;
			max_crossings  = 0;
			// All but first iteration consider all edges
		}

		cycleP = dllsP + max_crossings;
		for (edge1P = (edgeT *) cycleP->nextP; cycleP != (dllT *) edge1P; edge1P = next_edge1P) {
			next_edge1P = (edgeT *) edge1P->dll.nextP;
			assert(!edge1P->matched);
			assert(edge1P->blocks == max_crossings + 1);

#ifdef TEST_HILLCLIMB
			{
				int 	cnt;
				edgeT	*edge5P;
		
				for (edge4P = edgesP; edge4P < end_edgesP; ++edge4P) {
					code1aP = edge4P->code1P;
					code2aP = edge4P->code2P;
					if (edge4P->matched) {
						assert(code1aP->m_matchP == code2aP);
						assert(code2aP->m_matchP == code1aP);
						assert(!edge4P->blocks);
						continue;
					}
					assert(code1aP->m_matchP != code2aP);
					assert(code2aP->m_matchP != code1aP);
				
					cnt = 0;
					cross1P = edge4P->crossesP;
					for (end_cross1P = cross1P + edge4P->crosses; cross1P < end_cross1P; ++cross1P) {
						edge5P = edgesP + *cross1P;
						if (edge5P->matched) {
							++cnt;
					}	}
					assert(cnt == edge4P->blocks);
			}	}

			for (headP = dllsP; headP < end_dllsP; ++headP) {
				for (edge4P = (edgeT *) headP->nextP; headP != (dllT *) edge4P; edge4P = (edgeT *) edge4P->dll.nextP) {
					assert(edge4P->blocks == (headP - dllsP) + 1);
			}	}
#endif

			// Match1 is currently not an active edge
			// Compute the change in edges if all of the active edges crossing edge1P are made inactive
			// and edge1P is made active

#ifdef DEBUG_HILLCLIMB
			fprintf(stderr, "Considering activating %d<->%d\n",
								edge1P->code1P - start_code1P,
								edge1P->code2P - start_code2P);

			fprintf(stderr, "Initial add to reverse %p\n", edge1P);
#endif

			/* Edge 1 will become active if we commit this change which is
			 * an improvement of 1
			 */
			end_reversesP  = reversesP;
			change         = 1;
			/* Linked list of changed block numbers initially empty */
			changesP       = 0;
			end_unblockedP = unblockedP;

			cross1P = edge1P->crossesP;
			for (end_cross1P = cross1P + edge1P->crosses; cross1P < end_cross1P; ++cross1P) {
				/* Edge1 crosses edge2 */
				edge2P = edgesP + *cross1P;

				// Edge 1 will block this edge
				Increment(edge2P);

				if (!edge2P->matched) {
					continue;
				}
				// We have incremented all counts
				assert(edge2P->blocks1 == 1);
				// We must unactivate edge2P to permit activation of edge1
				Reverse(edge2P);
				// This is a match we loose
				--change;	
				cross2P = edge2P->crossesP;
				for (end_cross2P = cross2P + edge2P->crosses; cross2P < end_cross2P; ++cross2P) {
					// Because cross2 is inactive all edge3 crossing it are
					// no longer blocked by it
					edge3P = edgesP + *cross2P;
					assert(!edge3P->matched);
					Decrement(edge3P);
					if (edge3P->blocks1) {
						continue;
					}
					if (unblocked_maxP <= end_unblockedP) {
						at = max_unblocked;
						if (!max_unblocked) {
							max_unblocked   = 256;
						} else {
							max_unblocked <<= 1;
						}
						unblockedP = (unblockedT *) Xrealloc(unblockedP, max_unblocked * sizeof(unblockedT));
                        if (!unblockedP) {
                            fprintf(stderr, "Hillclimb: unblocked outofmemory\n");
                            max_unblocked = 0;
                            goto abort_reversals;
                        }

						unblocked_maxP = unblockedP + max_unblocked;
						end_unblockedP = unblockedP + at;
					}
					end_unblockedP->edgeP  = edge3P;
					++end_unblockedP;
			}	}

			/* We have now inactivated all edges crossing edge1 */
			assert(!(edge1P->blocks1));
			Reverse(edge1P);
			// Avoid allowing edge1P to match twice
			Increment(edge1P);

			for (;;) {
				for (unblocked1P = unblockedP; unblocked1P < end_unblockedP; ) {
					edge2P = unblocked1P->edgeP;
					if (edge2P->blocks1) {
						// No longer unblocked
						--end_unblockedP;
						*unblocked1P = *end_unblockedP;
						continue;
					} 
					edge2P->matched |= 2;
					++unblocked1P;
				}
				if (unblockedP == end_unblockedP) {
					break;
				}
				for (unblocked1P = unblockedP; unblocked1P < end_unblockedP; ++unblocked1P) {
					blocks  = 0;
					edge2P  = unblocked1P->edgeP;
					cross2P = edge2P->crossesP;
					for (end_cross2P = cross2P + edge2P->crosses; cross2P < end_cross2P; ++cross2P) {
						edge3P = edgesP + *cross2P;
						if (edge3P->matched & 2) {
							++blocks;
					}	}
					unblocked1P->blocks = blocks;
				}
				unblocked_bestP = unblocked1P = unblockedP;
				for (; unblocked1P < end_unblockedP; ++unblocked1P) {
					if (unblocked1P->blocks < unblocked_bestP->blocks) {
						unblocked_bestP = unblocked1P;
					}
					edge2P           = unblocked1P->edgeP;
					edge2P->matched &= ~2;
				}

				edge3P = unblocked_bestP->edgeP;
				--end_unblockedP;
				*unblocked_bestP = *end_unblockedP;

				assert(!edge3P->matched);
				assert(!edge3P->blocks1);

				// We will activate edge3P
				++change;	
				Reverse(edge3P);
				Increment(edge3P);	// Avoid choosing again

				cross3P = edge3P->crossesP;
				for (end_cross3P = cross3P + edge3P->crosses; cross3P < end_cross3P; ++cross3P) {
					// Edge3 activation blocks all things it crosses
					edge4P = edgesP + *cross3P;
					Increment(edge4P);
				}
			}
			if (change <= 0) {
				// This change does no good
				// Put all next_changeP pointers back to -1
abort_reversals:	
				for (edge2P = changesP; edge2P; edge2P = edge3P) {
					edge3P               = edge2P->next_changeP;
					edge2P->next_changeP = (edgeT *) -1;
				}
			} else {

#ifdef DEBUG_HILLCLIMB
				fprintf(stderr, "Found a change of %d in edge %d-%d\n",
							change,
							edge1P->code1P - start_code1P,
							edge1P->code2P - start_code2P);
#endif

				ret += change;

				// Apply all the reversals

				for (reverseP = reversesP; reverseP < end_reversesP; ++reverseP) {
					edge2P  = reverseP->edgeP;
					code1bP = edge2P->code1P;
					code2bP = edge2P->code2P;
#ifdef TEST_HILLCLIMB
					if (edge2P->matched != reverseP->matched) {
						reverseT	*reverse1P;

						for (reverse1P = reversesP; reverse1P <= reverseP; ++reverse1P) {
							edge3P = reverse1P->edgeP;
							fprintf(stderr, "[%d] %d<->%d ",
								edge3P - edgesP,
								edge3P->code1P - start_code1P,
								edge3P->code2P - start_code2P);
							if (reverse1P->matched) {
								fprintf(stderr, " unmatch\n");
							} else {
								fprintf(stderr, " match\n");
						}	}
						assert(edge2P->matched == reverseP->matched);
					}
#endif
					if (edge2P->matched) {
#ifdef DEBUG_HILLCLIMB
						fprintf(stderr, "Unmatch %d<->%d\n",
							code1bP - start_code1P,
							code2bP - start_code2P);
#endif
						// Unmatch this edge
						assert(code1bP->m_matchP == code2bP);
						assert(code2bP->m_matchP == code1bP);
						code1bP->m_matchP = 0;
						code2bP->m_matchP = 0;
						edge2P->matched   = 0;
						if (code1bP->m_jump.P > code1bP) {
							--forward_jumps;
						}
					} else {
#ifdef DEBUG_HILLCLIMB
						fprintf(stderr, "Match %d<->%d\n",
							code1bP - start_code1P,
							code2bP - start_code2P);
#endif
						// Match this edge
						assert(code1bP->m_matchP == 0);
						assert(code2bP->m_matchP == 0);
						code1bP->m_matchP = code2bP;
						code2bP->m_matchP = code1bP;
						edge2P->matched   = 1;
						if (code1bP->m_jump.P > code1bP) {
							++forward_jumps;
						}
						--(edge2P->blocks1); // Undo the blocking ++
					}
				}
				// Having applied all the reversals correct
				// The blocks crossed in all affected edges
				for (edge2P = changesP; edge2P; edge2P = edge3P) {
					edge3P         = edge2P->next_changeP;
					edge2P->blocks = blocks = edge2P->blocks1;
					if (blocks > g_max_blocks_seen) {
						g_max_blocks_seen = blocks;
					}
					edge2P->next_changeP = (edgeT *) -1;
					if ((headP = edge2P->dll.nextP)) {
						// Remove edge2 from dll's
						if (edge2P == next_edge1P) {
							next_edge1P = (edgeT *) headP;
						}
						headP->prevP             = edge2P->dll.prevP;
						edge2P->dll.prevP->nextP = headP;
					}
					if (blocks && blocks <= g_hillclimb) {
						// Add to the start of the right dll
						// We won't see it this iteration if adding it to
						// our cycle because before us in the cycle
						--blocks;
						headP                    = dllsP + blocks;
						edge2P->dll.prevP        = headP;
						edge2P->dll.nextP        = headP->nextP;
						headP->nextP             = (dllT *) edge2P;
						edge2P->dll.nextP->prevP = (dllT *) edge2P;
					} else {
						edge2P->dll.nextP        = 0;
						edge2P->dll.prevP        = 0;
					}
				}
			}


#ifdef TEST_HILLCLIMB
			for (headP = dllsP; headP < end_dllsP; ++headP) {
				for (edge4P = (edgeT *) headP->nextP; headP != (dllT *) edge4P; edge4P = (edgeT *) edge4P->dll.nextP) {
					assert(edge4P->blocks == (headP - dllsP) + 1);
			}	}
#endif
            if (hillclimb_terminate && hillclimb_terminate < time(0)) {
				method1P = function1P->m_methodP;
				method2P = function2P->m_methodP;

                fprintf(stderr, "* Hill-climbing aborted: timeout %d secs\n* ", g_hillclimb_timeout);
                dump_wchar(method1P->classP->P, stderr);
				fputc(':', stderr);
				dump_wchar(method1P->nameP->P, stderr);
				fputs(" .v.\n* ", stderr);
				dump_wchar(method2P->classP->P, stderr);
				fputc(':', stderr);
				dump_wchar(method2P->nameP->P, stderr);
				fputc('\n', stderr);    
                fprintf(stderr, "* Clone lths={%ld,%ld}, edges=%ld crosses=%u\n",
                    (long) (best_end1P - start_code1P),
                    (long) (best_end2P - start_code2P),
                    (long) (end_edgesP - edgesP),
                    crosses);

                goto terminate;
            }
	}	}

terminate:
	for (; !(best_end2P = best_end1P->m_matchP); --best_end1P);
#ifdef TEST_HILLCLIMB
	final_matches = 0;
	for (code1aP = start_code1P; code1aP <= best_end1P; ++code1aP) {
		if (code1aP->m_matchP) {
			++final_matches;
	}	}
	assert(final_matches - original_matches == ret);
#endif
done:
	g_hillclimb_elapse += time(0);
	return(ret);
}

typedef struct matchesS {
	struct matchesS *m_nextP;
	struct codeS	*m_codeP;
} matchesT;

typedef struct block_matchesS {
	struct block_matchesS	*m_nextP;
	matchesT				m_cache[8192];
} block_matchesT;

extern u4T				g_max_function_lth;

static int				g_matches_lth          = 0;
static matchesT			**g_matchesPP          = 0;
static block_matchesT	*g_head_block_matchesP = 0;		// Head of the linked list of blocks we can get matches from
static block_matchesT	*g_block_matchesP      = 0;		// The current block we are getting matches from
static matchesT	        *g_matches_cacheP      = 0;		// Next available match in the current block
static matchesT	        *g_matches_endP        = 0;		// End of available matches in the current block

static codeT *
detect_clone(void)
{
	static	codeT			*code1P, *code2P, *at_code1P, *at_code2P, *last_match1P;
	static	codeT			*max_end1P, *max_end2P;
	static	int				max_best, ret, weight, improved;
	static	int				original_matches, extra_matches, cancelled_matches;
	static  int				distance;
	static	block_matchesT	*next_blockP;
	static	matchesT		**matchesPP;

#ifdef DEBUG_DETECT
	static int debug_cnt = 0;

	fprintf(stderr, "%d Comparing %s(%d-%d) and %s(%d-%d)\n  %s\n  %s\n",  
		++debug_cnt,
		function1P->m_nameP,
		start_code1P - function_code1P,
		end_code1P - function_code1P,
		function2P->m_nameP,
		start_code2P - function_code2P, 
		end_code2P - function_code2P, 
		start_code1P->m_codeP, 
		start_code2P->m_codeP);
#endif
	start_code1P->m_matchP = start_code2P;
	start_code2P->m_matchP = start_code1P;
	last_match1P           = start_code1P;

	if (!same_code(start_code1P, start_code2P, &weight)) {
		goto fail;
	}
	max_best            = weight;
	max_end1P           = start_code1P;
	max_end2P           = start_code2P;

	start_clock = time(0);

	// Initially have to have a match
	best_end1P = code1P = start_code1P;
	best_end2P = code2P = start_code2P;
	original_matches    = 1;
	extra_matches       = 0;

extend_clone:
	improved = 0;
	for (; ++code1P < end_code1P && ++code2P < end_code2P;) {
		for (distance = 0; weight >= 0; weight += g_mismatch_cost) {
			at_code1P = code1P;
			at_code2P = code2P + distance;
			if (at_code2P >= end_code2P) {
				at_code1P += (at_code2P - end_code2P + 1);
				if (at_code1P >= end_code1P) {
					break;
				}
				at_code2P  = end_code2P - 1;
			} 
			while (at_code2P >= code2P && at_code1P < end_code1P) {
				if (same_code(at_code1P, at_code2P, &ret)) {
					goto found;
				}
				++at_code1P;
				--at_code2P;
			}
			++distance;
		}
		break;
found:	
		weight += ret;
		assert(weight >= 0);
		at_code1P->m_matchP = at_code2P;
		at_code2P->m_matchP = at_code1P;
		if (last_match1P < at_code1P) {
			last_match1P    = at_code1P;
		}
		improved        = 1;
		++original_matches;

		best_end1P  = at_code1P;
		best_end2P  = at_code2P;

		if (g_best_weight && (max_best <= weight)) {
			max_best  = weight;
			max_end1P = best_end1P;
			max_end2P = best_end2P;
		}
		code1P      = at_code1P;
		code2P      = at_code2P;
	}

	if (improved) {
		if (g_hillclimb > 0) {
			ret = hill_climb();
			if (ret > 0) {
				extra_matches += ret;
				weight += (ret * g_same_cost);
				code1P = best_end1P; 
				code2P = best_end2P;
				goto extend_clone;
	}	}	}
	
	if (g_best_weight) {
		weight     = max_best;
		best_end1P = max_end1P;
		best_end2P = max_end2P;
	}

	++best_end1P;
	++best_end2P;

	g_lth1     = best_end1P - start_code1P;
	g_lth2     = best_end2P - start_code2P;

	if (!permitted_clone()) {
		goto fail;
	}
	
	cancelled_matches = 0;
	for (code1P = start_code1P; code1P <= last_match1P; ++code1P) {
		at_code1P = code1P->m_jump.P;
		if (at_code1P <= code1P) {
			continue;
		}
		if (!(code2P = code1P->m_matchP)) {
			continue;
		}
		at_code2P = code2P->m_jump.P;
		if (at_code1P >= best_end1P) {
			if (at_code2P >= best_end2P) {
				// Both branch beyond end of clone
				// Might be turned into a return
				continue;
			}
		} else if (at_code2P < best_end2P) {
			if (at_code1P->m_matchP) {
				if (at_code2P == at_code1P->m_matchP) {
					// Both branch to the same matched instruction
					continue;
			}	}
			while (!((--at_code1P)->m_matchP));
			while (!((--at_code2P)->m_matchP));
			if (at_code1P->m_matchP == at_code2P) {
				// Both branch after the same matched instruction
				continue;
		}	}
		++cancelled_matches;
		if (code1P == start_code1P) {
			goto fail;
		}
		code1P->m_matchP = 0;
		code2P->m_matchP = 0;
	}

	g_original_matches  += original_matches;
	g_extra_matches     += extra_matches;
	g_cancelled_matches += cancelled_matches;
	end_clock = time(0);
#ifdef PROOF
	proof();
#endif

	report_clone();

	for (at_code1P = start_code1P; at_code1P <= last_match1P; ++at_code1P) {
		if ((at_code2P = at_code1P->m_matchP)) {

			if (g_matches_cacheP == g_matches_endP) {
				// Obtain another g_matches_cacheP memory area
				if (g_block_matchesP) {
					// If have further blocks use them
					next_blockP = g_block_matchesP->m_nextP;
				} else {
					next_blockP = 0;
				}
				if (!next_blockP) {
					next_blockP = (block_matchesT *) Xmalloc(sizeof(block_matchesT));
					if (!next_blockP) {
						outofmemory();
					}
					// Chain in newly created block
					next_blockP->m_nextP = 0;
					if (g_block_matchesP) {
						g_block_matchesP->m_nextP = next_blockP;
					} else {
						g_head_block_matchesP = next_blockP;
				}	}
				g_block_matchesP = next_blockP;

				// Give ourselves a new g_matches_cacheP
				g_matches_cacheP = g_block_matchesP->m_cache;
				g_matches_endP   = g_matches_cacheP + 8192;
			}
		
			// Add the fact that this line in function1P matched at_code2P
			g_matches_cacheP->m_codeP = at_code2P;
			matchesPP                 = g_matchesPP + (at_code1P - function_code1P);
			g_matches_cacheP->m_nextP = *matchesPP;
			*matchesPP                = g_matches_cacheP++;
	}	}
done:
	// Set any non-zero m_matchP pointers to zero
	for (; last_match1P >= start_code1P; --last_match1P) {
		if ((at_code2P = last_match1P->m_matchP)) {
			at_code2P->m_matchP    = 0;
			last_match1P->m_matchP = 0;
	}	}
	return(best_end2P);
fail:
	best_end2P = 0;
	goto done;
}

void
detect(void)
{
	static	Cfunction		*end_functions1P, *function3P;
	static	positionT		*position1P, *position2P;
	static	codeT			*skipP;
	static	matchesT		*matchesP, **matchesPP;
	static	time_t			clock_now;
	static	char			*P;
	static	int				lth;

	g_matches_lth     = sizeof(matchesT *) * g_max_function_lth;
	g_matchesPP       = (matchesT **)  Xmalloc(g_matches_lth);
	if (!g_matchesPP) {
		outofmemory();
	}

	for (source1P = Csource::g_source_headP; source1P; source1P = source1P->m_nextP) {
		for (class1P = source1P->m_classdefsP; class1P; class1P = class1P->m_nextP) {
			if (g_verbose) {
				clock_now = time(0);
				P         = ctime(&clock_now);
				P[19]     = 0;
				fprintf(stderr, "%s: ", P+11);
				dump_wchar(class1P->m_nameP->P, stderr);
				fputc('\n', stderr);
			}
			function1P = class1P->m_functionsP;
			for (end_functions1P = function1P + class1P->m_direct_functions + class1P->m_virtual_functions; function1P < end_functions1P; ++function1P) {
				start_code1P   = function_code1P = function1P->m_assemblersP;
				lth            = function1P->m_assemblers;
				function_end1P = start_code1P + lth;
				assert(lth <= (int) g_max_function_lth);
				memset(g_matchesPP, 0, lth * sizeof(matchesT *));
							
				for (; start_code1P < function_end1P; ++start_code1P) {
					if (g_match_lines) {
						if (start_code1P != function_code1P) {
							position1P = start_code1P[0].m_positionP;
							if (position1P) {
								position2P = start_code1P[-1].m_positionP;
								if (position2P) {
									if (position1P->line == position2P->line && position1P->source_fileP == position2P->source_fileP) {
										continue;
					}	}	}	}	}
										 
					for (start_code2P = start_code1P->m_hashP; start_code2P; ) {
						function2P      = start_code2P->m_functionP;
						class2P         = function2P->m_classP;
						source2P        = class2P->m_sourceP;

						// Detect subsequence within sequences already matched

						for (matchesPP = g_matchesPP + (start_code1P - function_code1P); (matchesP = *matchesPP); matchesPP = &(matchesP->m_nextP)) {
							if (matchesP->m_codeP == start_code2P) {
								*matchesPP = matchesP->m_nextP;
								goto advance;
						}	}

						function_code2P = function2P->m_assemblersP;
						function_end2P  = function_code2P + function2P->m_assemblers;
						if (function2P == function1P) {
							end_code2P = function_end1P;
							end_code1P = start_code2P;
						} else {
							end_code1P = function_end1P;
							end_code2P = function_end2P;
						}
						g_lth1 = end_code1P - start_code1P;
						g_lth2 = end_code2P - start_code2P;
						if (!permitted_clone()) {
							goto advance;
						}
						++g_comparisons;
						skipP         = detect_clone();
						start_code2P  = start_code2P->m_hashP;

						if (skipP) {
							for (; start_code2P; start_code2P = start_code2P->m_hashP) {
								function3P = start_code2P->m_functionP;
								if (function3P != function2P || start_code2P >= skipP) {
									break;
						}	}	}
						continue;
advance:				start_code2P = start_code2P->m_hashP;
				}	}

				// Set the matches back to the beginning
				g_block_matchesP = g_head_block_matchesP;
				if (g_block_matchesP) {
					g_matches_cacheP = g_block_matchesP->m_cache;
					g_matches_endP   = g_matches_cacheP + 8192;
				}
}	}	}	}
