#pragma once

// A fudge to allow us to document the dex file format

typedef u4T uleb128T;	// unsigned LEB128, variable-length (see below) 
typedef s4T	sleb128T;	// signed LEB128, variable-length (see below) 
typedef u4T uleb128p1T;	// unsigned LEB128 plus 1, variable-length (see below) 
#define REPEAT(X) 1
#define OFFSET(X,Y) Y
#define STRING_INDEX(X) X
#define TYPE_INDEX(X) X
#define METHOD_INDEX(X) X
#define PROTO_INDEX(X) X

// Throwable error

class Cdexerror {
public:
	const char	*P;

	Cdexerror(const char *messageP) 
	{
		P = messageP; 
	}

	static void error(const char *messageP);
	inline const char	*getError(void) { return P; }
};

// embedded in class_def_item and debug_info_item
// The constant NO_INDEX is used to indicate that an index value is absent.
// Note: This value isn't defined to be 0, because that is in fact typically a valid index.
// Also Note: The chosen value for NO_INDEX is representable as a single byte in the uleb128p1 encoding.

#define NO_INDEX 0xffffffff

// access_flags Definitions
// embedded in class_def_item, field_item, method_item, and InnerClass
// Bitfields of these flags are used to indicate the accessibility and overall properties of classes and class members.

//      Name                Value              For Classes (and InnerClass annotations)						For Fields										For Methods 
#define ACC_PUBLIC			0x1				// public: visible everywhere									public: visible everywhere						public: visible everywhere 
#define ACC_PRIVATE			0x2				// private: only visible to defining class	(*)					private: only visible to defining class			private: only visible to defining class 
#define ACC_PROTECTED		0x4				// protected: visible to package and subclasses (*)				protected: visible to package and subclasses	protected: visible to package and subclasses 
#define ACC_STATIC			0x8				// static: is not constructed with an outer this reference (*)	static: global to defining class				static: does not take a this argument 
#define ACC_FINAL			0x10			// final: not subclassable final: immutable after construction	final: not overridable 

#define ACC_SYNCHRONIZED	0x20			// synchronized: associated lock automatically acquired around call to this method. 
											// Note: This is only valid to set when ACC_NATIVE is also set. 
#define ACC_VOLATILE		0x40			// volatile: special access rules to help with thread safety   
#define ACC_BRIDGE			0x40			// bridge method, added automatically by compiler as a type-safe bridge 
#define ACC_TRANSIENT		0x80			// transient: not to be saved by default serialization   
#define ACC_VARARGS			0x80			// last argument should be treated as a "rest" argument by compiler 
#define ACC_NATIVE			0x100			// native: implemented in native code 
#define ACC_INTERFACE		0x200			// interface: multiply-implementable abstract class     
#define ACC_ABSTRACT		0x400			// abstract: not directly instantiable   abstract: unimplemented by this class 
#define ACC_STRICT			0x800			// strictfp: strict rules for floating-point arithmetic 
#define ACC_SYNTHETIC		0x1000			// not directly defined in source code not directly defined in source code not directly defined in source code 
#define ACC_ANNOTATION		0x2000			// declared as an annotation class     
#define ACC_ENUM			0x4000			// declared as an enumerated type declared as an enumerated value   
//(unused)					0x8000       
#define ACC_CONSTRUCTOR		0x10000			// constructor method (class or instance initializer) 
#define ACC_DECLARED_SYNCHRONIZED 0x20000	// declared synchronized. 
											// Note: This has no effect on execution (other than in reflection of this flag, per se).  

// (*) Only allowed on for InnerClass annotations, and must not ever be on in a class_def_item.


class Ctype_item {
public:
	// index into the type_ids list 
	u2T	TYPE_INDEX(type_idx);
};

// referenced from class_def_item and proto_id_item
// appears in the data section
// alignment: 4 bytes

class Ctype_list {
public:
	// size of the list, in entries
	u4T	size;
	// elements of the list 
	Ctype_item	list[REPEAT(size)];

	static void check_offset(u4T offset);
	static void fixup(u4T offset);
	static int  scan(u4T offset, stringT ***PPP, int load);
};

// map_item

class Cmap_item {

#define TYPE_HEADER_ITEM				0x0000 // size 0x70 
#define TYPE_STRING_ID_ITEM				0x0001 // size 0x04 
#define TYPE_TYPE_ID_ITEM				0x0002 // size 0x04 
#define TYPE_PROTO_ID_ITEM				0x0003 // size 0x0c 
#define TYPE_FIELD_ID_ITEM				0x0004 // size 0x08 
#define TYPE_METHOD_ID_ITEM				0x0005 // size 0x08 
#define TYPE_CLASS_DEF_ITEM				0x0006 // size 0x20 
#define TYPE_MAP_LIST					0x1000 // size 4 + (item.size * 12) 
#define TYPE_TYPE_LIST					0x1001 // size 4 + (item.size * 2) 
#define TYPE_ANNOTATION_SET_REF_LIST	0x1002 // size 4 + (item.size * 4) 
#define TYPE_ANNOTATION_SET_ITEM		0x1003 // size 4 + (item.size * 4) 
#define TYPE_CLASS_DATA_ITEM			0x2000 // implicit; must parse 
#define TYPE_CODE_ITEM					0x2001 // implicit; must parse 
#define TYPE_STRING_DATA_ITEM			0x2002 // implicit; must parse 
#define TYPE_DEBUG_INFO_ITEM			0x2003 // implicit; must parse 
#define TYPE_ANNOTATION_ITEM			0x2004 // implicit; must parse 
#define TYPE_ENCODED_ARRAY_ITEM			0x2005 // implicit; must parse 
#define TYPE_ANNOTATIONS_DIRECTORY_ITEM	0x2006 // implicit; must parse 

	// type of the items; see table below 
	u2T	type;
	// (unused) 
	u2T	unused;
	// count of the number of items to be found at the indicated offset
	u4T	size;
	// offset from the start of the file to the items in question 
	u4T	offset;
};

// map_list
// appears in the data section
// referenced from header_item
// alignment: 4 bytes
// This is a list of the entire contents of a file, in order. 
// It contains some redundancy with respect to the header_item but is intended to be an easy form to use to iterate over an entire file. 
// A given type may appear at most once in a map, but there is no restriction on what order types may appear in, other than the restrictions implied by the rest of the format 
// (e.g., a header section must appear first, followed by a string_ids section, etc.). 
// Additionally, the map entries must be ordered by initial offset and must not overlap.

class Cmap_list {
public:
	// size of the list, in entries 
	u4T	size;
	// elements of the list 
	Cmap_item	list[REPEAT(size)];
};

// string_data_item
// appears in the data section
// alignment: none (byte-aligned)

class Cstring_data_item {
public:
	// size of this string, in UTF-16 code units (which is the "string length" in many systems). 
	// That is, this is the decoded length of the string. 
	// (The encoded length is implied by the position of the 0 byte.) 
	uleb128T	utf16_size;
	// a series of MUTF-8 code units (a.k.a. octets, a.k.a. bytes) followed by a byte of value 0. 
	// See "MUTF-8 (Modified UTF-8) Encoding" above for details and discussion about the data format. 
	// Note: It is acceptable to have a string which includes (the encoded form of) UTF-16 surrogate code units 
	// (that is, U+d800  U+dfff) either in isolation or out-of-order with respect to the usual encoding of Unicode into UTF-16. 
	// It is up to higher-level uses of strings to reject such invalid encodings, if appropriate.
	u1T			data[REPEAT(0)];
};

// string_id_item
// appears in the string_ids section
// alignment: 4 bytes

class Cstring_id_item {
public:
	// offset from the start of the file to the string data for this item. 
	// The offset should be to a location in the data section, and the data should be in the format specified by "string_data_item" above. 
	// There is no alignment requirement for the offset.  
	u4T	OFFSET(Cstring_data_item, string_data_off);

	static void check_offset(u4T size, u4T offset);
	static void fixup_all(u4T size, u4T offset);
	static void scan_all(u4T size, u4T offset);
};

// type_id_item
// appears in the type_ids section
// alignment: 4 bytes

class Ctype_id_item {
public:
	// index into the string_ids list for the descriptor string of this type. 
	//The string must conform to the syntax for TypeDescriptor, defined above. 
	u4T STRING_INDEX(descriptor_idx);

	static void check_offset(u4T size, u4T offset);
	static void fixup_all(u4T size, u4T offset);
	static void scan_all(u4T size, u4T offset);
};

// proto_id_item
// appears in the proto_ids section
// alignment: 4 bytes

// method prototype identifiers list. 
// These are identifiers for all prototypes referred to by this file. 
// This list must be sorted in return-type (by type_id index) major order, and then by arguments (also by type_id index).

class Cproto_id_item {
public:
	// index into the string_ids list for the short-form descriptor string of this prototype. 
	// The string must conform to the syntax for ShortyDescriptor, defined above, and must correspond to the return type and parameters of this item.  
	u4T	STRING_INDEX(shorty_idx);
	// index into the type_ids list for the return type of this prototype  
	u4T	TYPE_INDEX(return_type_idx);
	// offset from the start of the file to the list of parameter types for this prototype, or 0 if this prototype has no parameters. 
	// This offset, if non-zero, should be in the data section, and the data there should be in the format specified by "type_list" above. 
	// Additionally, there should be no reference to the type void in the list.  
	u4T	OFFSET(Ctype_list, parameters_off);

	static void check_offset(u4T size, u4T offset);
	static void fixup_all(u4T size, u4T offset);
	static void scan_all(u4T size, u4T offset);
};

// field_id_item
// appears in the field_ids section
// alignment: 4 bytes

class Cfield_id_item {
public:
	// index into the type_ids list for the definer of this field. This must be a class type, and not an array or primitive type. 
	u2T	TYPE_INDEX(class_idx);
	// index into the type_ids list for the type of this field  
	u2T	TYPE_INDEX(type_idx);
	// index into the string_ids list for the name of this field. The string must conform to the syntax for MemberName, defined above.
	u4T STRING_INDEX(name_idx);

	static void check_offset(u4T size, u4T offset);
	static void fixup_all(u4T size, u4T offset);
	static void scan_all(u4T size, u4T offset);
};

// method_id_item
// appears in the method_ids section
// alignment: 4 bytes

class Cmethod_id_item {
public:
	// index into the type_ids list for the definer of this method. This must be a class or array type, and not a primitive type.
	u2T	TYPE_INDEX(class_idx);
	// index into the proto_ids list for the prototype of this method  
	u2T	PROTO_INDEX(proto_idx);
	// index into the string_ids list for the name of this method. The string must conform to the syntax for MemberName, defined above. 
	u4T	STRING_INDEX(name_idx);

	static void check_offset(u4T size, u4T offset);
	static void fixup_all(u4T size, u4T offset);
	static void scan_all(u4T size, u4T offset);
};

class Cencoded_field {
public:
	// index into the field_ids list for the identity of this field 
	// (includes the name and descriptor), 
	// represented as a difference from the index of previous element in the list. 
	// The index of the first element in a list is represented directly.  
	uleb128T field_idx_diff;
	// access flags for the field (public, final, etc.). See "access_flags Definitions" for details. 
	uleb128T access_flags;

	static void fixup(u1T **PP);
};

class Cencoded_type_addr_pair {
public:
	// index into the type_ids list for the type of the exception to catch 
	uleb128T TYPE_INDEX(type_idx);
	// bytecode address of the associated exception handler 
	uleb128T addr;
};

class Cencoded_catch_handler {
public:
	// number of catch types in this list. 
	// If non-positive, then this is the negative of the number of catch types, and the catches are followed by a catch-all handler. 
	// For example: A size of 0 means that there is a catch-all but no explicitly typed catches. 
	// A size of 2 means that there are two explicitly typed catches and no catch-all. 
	// And a size of -1 means that there is one typed catch along with a catch-all.  
	sleb128T size;
	// stream of abs(size) encoded items, one for each caught type, in the order that the types should be tested.
	Cencoded_type_addr_pair handlers[REPEAT(abs(size))];
	// (optional) bytecode address of the catch-all handler. This element is only present if size is non-positive.  
	uleb128T catch_all_addr;
};

class Cencoded_catch_handler_list {
public:
	// size of this list, in entries
	uleb128T size;
	// actual list of handler lists, represented directly (not as offsets), and concatenated sequentially 
	Cencoded_catch_handler	list[REPEAT(handlers_size)];
};

class Ctry_item {
public:
	// start address of the block of code covered by this entry. The address is a count of 16-bit code units to the start of the first covered instruction.  
	u4T	start_addr; 
	// number of 16-bit code units covered by this entry. The last code unit covered (inclusive) is start_addr + insn_count - 1.  
	u2T insn_count;
	// offset in bytes from the start of the associated encoded handler data to the catch_handler_item for this entry  
	u2T OFFSET(Cencoded_catch_handler, handler_off);

	void fixup();
};

class Cdebug_info_item {
public:
	// line_start uleb128 the initial value for the state machine's line register. Does not represent an actual positions entry.  
	u4T	line_start;
	// parameters_size uleb128 the number of parameter names that are encoded. There should be one per method parameter, excluding an instance method's this, if any.  
	u4T parameters_size;
	// parameter_names uleb128p1[parameters_size] string index of the method parameter name. An encoded value of NO_INDEX indicates that no name is available for the associated parameter. The type descriptor and signature are implied from the method descriptor and signature.  
	u4T	parameter_names[1];

	static void scan(u4T offset);
};

// code_item
// referenced from method_item
// appears in the data section
// alignment: 4 bytes

class Ccode_item {
public:
	// the number of registers used by this code
	u2T registers_size;
	// the number of words of incoming arguments to the method that this code is for 
	u2T	ins_size;
	// the number of words of outgoing argument space required by this code for method invocation
	u2T	outs_size;
	// the number of try_items for this instance. If non-zero, then these appear as the tries array just after the insns in this instance.  
	u2T	tries_size;
	// offset from the start of the file to the debug info (line numbers + local variable info) sequence for this code, or 0 if there simply is no information. 
	// The offset, if non-zero, should be to a location in the data section. 
	// The format of the data is specified by "debug_info_item" below.  
	u4T	debug_info_off;
	// size of the instructions list, in 16-bit code units 
	u4T	insns_size;
	// actual array of bytecode. The format of code in an insns array is specified by the companion document "Bytecode for the Dalvik VM". Note that though this is defined as an array of ushort, there are some internal structures that prefer four-byte alignment. Also, if this happens to be in an endian-swapped file, then the swapping is only done on individual ushorts and not on the larger internal structures.  
	// (optional) = 0 two bytes of padding to make tries four-byte aligned. 
	// This element is only present if tries_size is non-zero and insns_size is odd.  	
	u2T	insns[REPEAT((insns_size+1)&~1)];
	// (optional) array indicating where in the code exceptions may be caught and how to handle them. Elements of the array must be non-overlapping in range and in order from low to high address. This element is only present if tries_size is non-zero.  
	Ctry_item tries[REPEAT(tries_size)];
	// (optional) bytes representing a list of lists of catch types and associated handler addresses. 
	// Each try_item has a byte-wise offset into this structure. 
	// This element is only present if tries_size is non-zero.  
	Cencoded_catch_handler_list handlers;

	static void check_offset(u4T offset);
	static void fixup(u4T offset);
	static void scan(u4T offset);
};

class Cencoded_method {
public:
	// index into the method_ids list for the identity of this method (includes the name and descriptor), 
	// represented as a difference from the index of previous element in the list. 
	// The index of the first element in a list is represented directly.  
	uleb128T	method_idx_diff;
	// access flags for the method (public, final, etc.). 
	// See "access_flags Definitions" for details.  
	uleb128T	access_flags;
	// offset from the start of the file to the code structure for this method, or 0 if this method is either abstract or native. 
	// The offset should be to a location in the data section. 
	// The format of the data is specified by "code_item" above.  
	uleb128T	code_off;

	static void fixup(u1T **PP);
};

// class_data_item
// referenced from class_def_item
// appears in the data section
// alignment: none (byte-aligned)
// Note: All elements' field_ids and method_ids must refer to the same defining class.

class Cclass_data_item {
public:
	// the number of static fields defined in this item 
	uleb128T static_fields_size;
	// the number of instance fields defined in this item 
	uleb128T instance_fields_size;
	// direct_methods_size uleb128 the number of direct methods defined in this item 
	uleb128T direct_methods_size;
	// the number of virtual methods defined in this item 
	uleb128T virtual_methods_size;
	// the defined static fields, represented as a sequence of encoded elements. 
	// The fields must be sorted by field_idx in increasing order.  
	Cencoded_field	static_fields[REPEAT(static_fields_size)];
	// the defined instance fields, represented as a sequence of encoded elements. The fields must be sorted by field_idx in increasing order.  
	Cencoded_field	instance_fields[REPEAT(instance_fields_size)];
	// the defined direct (any of static, private, or constructor) methods, represented as a sequence of encoded elements. The methods must be sorted by method_idx in increasing order.  
	Cencoded_method	direct_methods[REPEAT(direct_methods_size)];
	// the defined virtual (none of static, private, or constructor) methods, represented as a sequence of encoded elements. This list should not include inherited methods unless overridden by the class that this item represents. The methods must be sorted by method_idx in increasing order.  
	Cencoded_method virtual_methods[REPEAT(virtual_methods_size)];

	static void fixup(u4T offset);
	static void scan(u4T offset);
};

// Dex annotations

/*	System Annotations

	System annotations are used to represent various pieces of reflective information about classes (and methods and fields). 
	This information is generally only accessed indirectly by client (non-system) code.

	System annotations are represented in .dex files as annotations with visibility set to VISIBILITY_SYSTEM. 

	dalvik.annotation.AnnotationDefault
	appears on methods in annotation interfaces
	An AnnotationDefault annotation is attached to each annotation interface which wishes to indicate default bindings.
	the default bindings for this annotation, represented as an annotation of this type. 
	
	// The annotation need not include all names defined by the annotation; missing names simply do not have defaults.  
	Annotation value;

	dalvik.annotation.EnclosingClass
	appears on classes
	An EnclosingClass annotation is attached to each class which is either defined as a member of another class, per se, or is anonymous but not defined within a method body 
	(e.g., a synthetic inner class). 
	Every class that has this annotation must also have an InnerClass annotation. 
	Additionally, a class may not have both an EnclosingClass and an EnclosingMethod annotation.

	// the class which most closely lexically scopes this class 
	Class value; 

	dalvik.annotation.EnclosingMethod
	appears on classes
	An EnclosingMethod annotation is attached to each class which is defined inside a method body. 
	Every class that has this annotation must also have an InnerClass annotation. 
	Additionally, a class may not have both an EnclosingClass and an EnclosingMethod annotation.

	// the method which most closely lexically scopes this class 
	Method value;

	dalvik.annotation.InnerClass
	appears on classes
	An InnerClass annotation is attached to each class which is defined in the lexical scope of another class's definition. 
	Any class which has this annotation must also have either an EnclosingClass annotation or an EnclosingMethod annotation.

	// the originally declared simple name of this class (not including any package prefix). If this class is anonymous, then the name is null. 
	String name;
	// the originally declared access flags of the class (which may differ from the effective flags because of a mismatch between the execution models of the source language and target virtual machine)  
	int accessFlags;

	dalvik.annotation.MemberClasses
	appears on classes
	A MemberClasses annotation is attached to each class which declares member classes. 
	(A member class is a direct inner class that has a name.)

	// array of the member classes 
	Class value[];

	dalvik.annotation.Signature
	appears on classes, fields, and methods
	A Signature annotation is attached to each class, field, or method which is defined in terms of a more complicated type than is representable by a type_id_item. 
	The .dex format does not define the format for signatures; it is merely meant to be able to represent whatever signatures a source language requires for successful implementation of that language's semantics. 
	As such, signatures are not generally parsed (or verified) by virtual machine implementations. 
	The signatures simply get handed off to higher-level APIs and tools (such as debuggers). 
	Any use of a signature, therefore, should be written so as not to make any assumptions about only receiving valid signatures, explicitly guarding itself against the possibility of coming across a syntactically invalid signature.

	Because signature strings tend to have a lot of duplicated content, a Signature annotation is defined as an array of strings, 
	where duplicated elements naturally refer to the same underlying data, and the signature is taken to be the concatenation of all the strings in the array. 
	There are no rules about how to pull apart a signature into separate strings; that is entirely up to the tools that generate .dex files.

	// the signature of this class or member, as an array of strings that is to be concatenated together
	String value[];

	dalvik.annotation.Throws
	appears on methods
	A Throws annotation is attached to each method which is declared to throw one or more exception types.

	// the array of exception types thrown 
	Class value[];
*/


// embedded in annotation_element and encoded_array_item
// An encoded_value is an encoded piece of (nearly) arbitrary hierarchically structured data. 
// The encoding is meant to be both compact and straightforward to parse.

class Cencoded_value {
public:
	// (value_arg << 5) | value_type byte indicating the type of the immediately subsequent value 
	// along with an optional clarifying argument in the high-order three bits. 
	// See below for the various value definitions. 
	// In most cases, value_arg encodes the length of the immediately-subsequent value in bytes, as (size - 1), 
	// e.g., 0 means that the value requires one byte, and 7 means it requires eight bytes; however, there are exceptions as noted below.  
	u1T	value_arg_type;
	// bytes representing the value, variable in length and interpreted differently for different value_type bytes, though always little-endian. 
	// See the various value definitions below for details.  
	u1T	value[1];

	static void scan(u1T **PP, annotation_valueT *targetP, u1T **memoryPP, int load);
};

class Cencoded_array {
public:
	// number of elements in the array
	uleb128T size;
	// a series of size encoded_value byte sequences in the format specified by this section, concatenated sequentially.  
	Cencoded_value	values[REPEAT(size)];
};

class Cannotation_element {
public:
	// element name, represented as an index into the string_ids section. The string must conform to the syntax for MemberName, defined above. 
	uleb128T		name_idx;
	// element value
	Cencoded_value	value;
};

class Cencoded_annotation {
public:
	// type of the annotation. This must be a class (not array or primitive) type.  
	uleb128T type_idx;
	// number of name-value mappings in this annotation
	uleb128T size;
	// elements of the annotataion, represented directly in-line (not as offsets). Elements must be sorted in increasing order by string_id index.  
	Cannotation_element	elements[REPEAT(size)];

	static void scan(u1T **PP, annotationT *targetP, u1T **memoryPP, int load);
} ;

// annotation_item
// referenced from annotation_set_item
// appears in the data section
// alignment: none (byte-aligned)

class Cannotation_item {
public:
	// intended visibility of this annotation (see below) 
	u4T	visibility;
#define VISIBILITY_BUILD	0x00 // intended only to be visible at build time (e.g., during compilation of other code)  
#define VISIBILITY_RUNTIME	0x01 // intended to visible at runtime 
#define VISIBILITY_SYSTEM	0x02 // intended to visible at runtime, but only to the underlying system (and not to regular user code 
	// encoded annotation contents, in the format described by "encoded_annotation Format" under "encoded_value Encoding" above.
	Cencoded_annotation	annotation;

	static void scan(u4T offset, annotation_itemT *targetP, u1T **memoryPP, int load);
};

class Cannotation_off_item {
public:
	// offset from the start of the file to an annotation. 
	// The offset should be to a location in the data section, and the format of the data at that location is specified by "annotation_item" below.  
	u4T	annotation_off;

	void scan(annotation_itemT *targetP, u1T **memoryPP, int load);
};

// annotation_set_item
// referenced from annotations_directory_item, field_annotations_item, method_annotations_item, and annotation_set_ref_item
// appears in the data section
// alignment: 4 bytes

class Cannotation_set_item {
public:
	// size of the set, in entries
	u4T	size;
	// elements of the set. The elements must be sorted in increasing order, by type_idx
	Cannotation_off_item entries[REPEAT(size)];

	static void check_offset(u4T offset);
	static void fixup(u4T offset);
	static void scan(u4T offset, annotation_setT *annotation_setP, u1T **memoryPP, int load);
};

// annotation_set_ref_item
// referenced from parameter_annotations_item
// appears in the data section
// alignment: 4 bytes

class Cannotation_set_ref_item {
public:
	// offset from the start of the file to the referenced annotation set or 0 if there are no annotations for this element. 
	// The offset, if non-zero, should be to a location in the data section. 
	// The format of the data is specified by "annotation_set_item" below.  
	u4T	OFFSET(Cannotation_set_item, annotations_off);

	void fixup(void);
};

// annotation_set_ref_list
// referenced from parameter_annotations_item
// appears in the data section
// alignment: 4 bytes

class Cannotation_set_ref_list {
public:
	 // size of the list, in entries 
	u4T size;
	// elements of the list 
	Cannotation_set_ref_item	list[REPEAT(size)];

	static void check_offset(u4T offset);
	static void fixup(u4T offset);
};

class Cfield_annotation {
public:
	// index into the field_ids list for the identity of the field being annotated 
	u4T	field_idx;
	// offset from the start of the file to the list of annotations for the field. 
	// The offset should be to a location in the data section. 
	// The format of the data is specified by "annotation_set_item" below.  
	u4T	OFFSET(Cannotion_set_item, annotations_off);

	void fixup(void);
	void scan(annotation_setT *targetP, u1T **memoryPP, int load);
};

class Cmethod_annotation {
public:
	// index into the method_ids list for the identity of the method being annotated  
	u4T	method_idx;
	// offset from the start of the file to the list of annotations for the method. 
	// The offset should be to a location in the data section. 
	// The format of the data is specified by "annotation_set_item" below.  
	u4T	annotations_off;

	void fixup(void);
	void scan(annotation_setT *targetP, u1T **memoryPP, int load);
};

class Cparameter_annotation {
public:
	// index into the method_ids list for the identity of the method whose parameters are being annotated 
	u4T	METHOD_INDEX(method_idx);
	// offset from the start of the file to the list of annotations for the method parameters. 
	// The offset should be to a location in the data section. 
	// The format of the data is specified by "annotation_set_ref_list" below.  
	u4T	OFFSET(Cannotation_set_ref_list, annotations_off);

	void	fixup(void);
	void	scan(annotation_parameterT **targetPP, u1T **memoryPP, int load);
};

// annotations_directory_item
// referenced from class_def_item
// appears in the data section
// alignment: 4 bytes

class Cannotations_directory_item {
public:
	// offset from the start of the file to the annotations made directly on the class, or 0 if the class has no direct annotations. 
	// The offset, if non-zero, should be to a location in the data section. 
	// The format of the data is specified by "annotation_set_item" above.  
	u4T	class_annotations_off;
	// count of fields annotated by this item 
	u4T	fields_size;
	// count of methods annotated by this item 
	u4T	methods_size;
	// count of method parameter lists annotated by this item 
	u4T	parameters_size;
	// (optional) list of associated field annotations. The elements of the list must be sorted in increasing order, by field_idx.  
	Cfield_annotation	field_annotations[REPEAT(fields_size)];
	// (optional) list of associated method annotations. The elements of the list must be sorted in increasing order, by method_idx.  
	Cmethod_annotation	method_annotations[REPEAT(method_size)];
	// (optional) list of associated method parameter annotations. The elements of the list must be sorted in increasing order, by method_idx.  
	Cparameter_annotation	parameter_annotations[REPEAT(parameters_size)];

	static void check_offset(u4T offset);
	static void fixup(u4T offset);
	static void scan(annotationsT *targetP, u4T offset);
};

// class_def_item
// appears in the class_defs section
// alignment: 4 bytes

// class definitions list. 
// The classes must be ordered such that a given class's superclass and implemented interfaces appear in the list earlier than the referring class. 

class Cclass_def_item {
public:
	// index into the type_ids list for this class. This must be a class type, and not an array or primitive type. 
	u4T	class_idx;
	// access flags for the class (public, final, etc.). See "access_flags Definitions" for details.
	u4T access_flags;
	// index into the type_ids list for the superclass, or the constant value NO_INDEX if this class has no superclass (i.e., it is a root class such as Object). 
	// If present, this must be a class type, and not an array or primitive type.  
	u4T	superclass_idx;
	// offset from the start of the file to the list of interfaces, or 0 if there are none. 
	// This offset should be in the data section, and the data there should be in the format specified by "type_list" below. 
	// Each of the elements of the list must be a class type (not an array or primitive type), and there must not be any duplicates.  
	u4T interfaces_off;
	// index into the string_ids list for the name of the file containing the original source for (at least most of) this class, 
	// or the special value NO_INDEX to represent a lack of this information. 
	// The debug_info_item of any given method may override this source file, but the expectation is that most classes will only come from one source file.  
	u4T	source_file_idx;
	// offset from the start of the file to the annotations structure for this class, or 0 if there are no annotations on this class. 
	// This offset, if non-zero, should be in the data section, and the data there should be in the format specified by "annotations_directory_item" below, 
	// with all items referring to this class as the definer.  
	u4T annotations_off;
	// offset from the start of the file to the associated class data for this item, or 0 if there is no class data for this class. 
	// (This may be the case, for example, if this class is a marker interface.) 
	// The offset, if non-zero, should be in the data section, and the data there should be in the format specified by "class_data_item" below, with all items referring to this class as the definer.  
	u4T class_data_off;
	//	offset from the start of the file to the list of initial values for static fields, 
	// or 0 if there are none (and all static fields are to be initialized with 0 or null). 
	// This offset should be in the data section, and the data there should be in the format specified by "encoded_array_item" below. 
	// The size of the array must be no larger than the number of static fields declared by this class, and the elements correspond to the static fields in the same order as declared in the corresponding field_list. 
	// The type of each array element must match the declared type of its corresponding field. 
	// If there are fewer elements in the array than there are static fields, then the leftover fields are initialized with a type-appropriate 0 or null.  
	u4T	static_values_off;

	static void check_offset(u4T size, u4T offset);
	static void fixup_all(u4T size, u4T offset);
	static void scan_all(u4T size, u4T offset);
};

#define DEX_FILE_MAGIC "dex\n035\0"

// header_item
// appears in the header section
// alignment: 4 bytes

class Cheader_item {
public:
	// magic ubyte[8] = DEX_FILE_MAGIC magic value. See discussion above under "DEX_FILE_MAGIC" for more details.  
	char	magic[8];
	// checksum uint adler32 checksum of the rest of the file (everything but magic and this field); used to detect file corruption  
	u4T		checksum;
	// signature ubyte[20] SHA-1 signature (hash) of the rest of the file (everything but magic, checksum, and this field); used to uniquely identify files  
	char	signature[20];
	// file_size uint size of the entire file (including the header), in bytes
	u4T 	file_size;
	// header_size uint = 0x70 size of the header (this entire section), in bytes. 
	// This allows for at least a limited amount of backwards/forwards compatibility without invalidating the format.
	u4T 	header_size;
	// endian_tag uint = ENDIAN_CONSTANT endianness tag. 
	// Although the standard .dex format is little-endian, implementations may choose to perform byte-swapping. 
	// Should an implementation come across a header whose endian_tag is REVERSE_ENDIAN_CONSTANT instead of ENDIAN_CONSTANT, 
	// it would know that the file has been byte-swapped from the expected form.

#define ENDIAN_CONSTANT			0x12345678
#define REVERSE_ENDIAN_CONSTANT 0x78563412

	u4T 	endian_tag;
	// link_size uint size of the link section, or 0 if this file isn't statically linked 
	u4T 	link_size;
	// link_off uint offset from the start of the file to the link section, 
	// or 0 if link_size == 0. The offset, if non-zero, should be to an offset into the link_data section. 
	// The format of the data pointed at is left unspecified by this document; this header 
	// field (and the previous) are left as hooks for use by runtime implementations. 
	u4T 	link_off;
	// map_off uint offset from the start of the file to the map item, or 0 if this file 
	// has no map. The offset, if non-zero, should be to an offset into the data section, 
	// and the data should be in the format specified by "map_list" below.
	u4T 	OFFSET(Cmap_list, map_off);
	// string_ids_size uint count of strings in the string identifiers list 
	u4T 	string_ids_size;
	// string_ids_off uint offset from the start of the file to the string identifiers 
	// list, or 0 if string_ids_size == 0 (admittedly a strange edge case). The offset, 
	// if non-zero, should be to the start of the string_ids section.  
	u4T 	OFFSET(Cstring_ids_item, string_ids_off);
	// type_ids_size uint count of elements in the type identifiers list 
	u4T 	type_ids_size;
	// type_ids_off uint offset from the start of the file to the type identifiers list, 
	// or 0 if type_ids_size == 0 (admittedly a strange edge case). The offset, if 
	// non-zero, should be to the start of the type_ids section.  
	u4T 	OFFSET(Ctype_ids_item, type_ids_off);
	// proto_ids_size uint count of elements in the prototype identifiers list 
	u4T 	proto_ids_size;
	// proto_ids_off uint offset from the start of the file to the prototype identifiers 
	// list, or 0 if proto_ids_size == 0 (admittedly a strange edge case). The offset, 
	// if non-zero, should be to the start of the proto_ids section.  
	u4T 	proto_ids_off;
	// field_ids_size uint count of elements in the field identifiers list 
	u4T 	field_ids_size;
	// field_ids_off uint offset from the start of the file to the field identifiers list, 
	// or 0 if field_ids_size == 0. The offset, if non-zero, should be to the start of the 
	// field_ids section. 
	u4T 	field_ids_off;
	// method_ids_size uint count of elements in the method identifiers list 
	u4T 	method_ids_size;
	// method_ids_off uint offset from the start of the file to the method identifiers 
	// list, or 0 if method_ids_size == 0. The offset, if non-zero, should be to the 
	// start of the method_ids section. 
	u4T 	method_ids_off;
	// class_defs_size uint count of elements in the class definitions list
	u4T 	class_defs_size;
	// class_defs_off uint offset from the start of the file to the class definitions list,
	// or 0 if class_defs_size == 0 (admittedly a strange edge case). The offset, if 
	// non-zero, should be to the start of the class_defs section. 	
	u4T 	class_defs_off;
	// data_size uint Size of data section in bytes. Must be an even multiple of sizeof(uint). 
	u4T 	data_size;
	// data_off uint offset from the start of the file to the start of the data section.
	u4T 	data_off;

	static void fixup(void);
	static void scan(void);
};

class Cdex_file_layout {
public:
	// the header 
	Cheader_item	header;	
	// string identifiers list. 
	// These are identifiers for all the strings used by this file, either for internal naming (e.g., type descriptors) or as constant objects referred to by code.
	// This list must be sorted by string contents, using UTF-16 code point values (not in a locale-sensitive manner).  
	Cstring_id_item	string_ids[REPEAT(0)];
	// type identifiers list. 
	// These are identifiers for all types (classes, arrays, or primitive types) referred to by this file, whether defined in the file or not. 
	// This list must be sorted by string_id index.  
	Ctype_id_item	type_ids[REPEAT(0)];
	// method prototype identifiers list. 
	// These are identifiers for all prototypes referred to by this file. 
	// This list must be sorted in return-type (by type_id index) major order, and then by arguments (also by type_id index).  
	Cproto_id_item	proto_ids[REPEAT(0)];
	// field identifiers list. 
	// These are identifiers for all fields referred to by this file, whether defined in the file or not. 
	// This list must be sorted, where the defining type (by type_id index) is the major order, field name (by string_id index) is the intermediate order, and type (by type_id index) is the minor order.  
	Cfield_id_item	field_ids[REPEAT(0)];
	// method identifiers list. 
	// These are identifiers for all methods referred to by this file, whether defined in the file or not. 
	// This list must be sorted, where the defining type (by type_id index) is the major order, method name (by string_id index) is the intermediate order, and method prototype (by proto_id index) is the minor order.  
	Cmethod_id_item	method_ids[REPEAT(0)];
	// class definitions list. 
	// The classes must be ordered such that a given class's superclass and implemented interfaces appear in the list earlier than the referring class.  
	Cclass_def_item class_defs[REPEAT(0)];
	// data area, containing all the support data for the tables listed above. Different items have different alignment requirements, and padding bytes are inserted before each item if necessary to achieve proper alignment.  
	u1T	data[REPEAT(0)];
	// data used in statically linked files. The format of the data in this section is left unspecified by this document; this section is empty in unlinked files, and runtime implementations may use it as they see fit.  
	u1T	link_data[REPEAT(0)];
};

