#include "typedef.h"

#include <assert.h>
#include <stdio.h>
#include <string.h>
#ifndef WIN32
#include <unistd.h>
#endif
#include <stdlib.h>

#include "xmalloc.h"
#include "util.h"
#include "object.h"
#include "opcode.h"
#include "ta.h"
#include "file.h"
#include "code.h"
#include "clone.h"
#include "source.h"
#include "class.h"
#include "function.h"

/*
static void trap(void)
{
}
*/

static int
relativeToShort(double value)
{
	assert(value >= 0.0);
	assert(value <= 1.0);
	return (int) ((value - 0.5) * 65534.0);
}

extern FILE *taF;
extern FILE	*tempF;
extern FILE	*stdoutF;

/* static */ void
Cfunction::dump_descriptor(const u2T *descriptorP)
{
	const	char *P;
	u2T		c, c1;
	int		dimension = 0;
	int		parameter = 0;
	
//	dumpUnicode(stdoutF, descriptorP);
//	fputs("->", stdoutF);

	for (;;) {
		switch (c = *descriptorP++) {
		case 0:
			return;
		case '[':
			++dimension;
			continue;
			break;
		case 'L':
			if (parameter > 1) {
				fputs(", ", stdoutF);
			}
			for (; (c1 = *descriptorP) && c1 != ';'; ++descriptorP) {
				if (c1 == '/') {
					c1 = '.';
				}
				fputc(c1, stdoutF);
			}
			goto show_array;
		case '(':
			fputc('(', stdoutF);
			parameter = 1;
			continue;
		case ')':
			fputs(") returns ", stdoutF);
			parameter = 0;
			continue;
		case 'B':
			P = "byte";
			break;
		case 'C':
			P = "char";
			break;
		case 'D':
			P = "double";
			break;
		case 'F':
			P = "float";
			break;
		case 'I':
			P = "int";
			break;
		case 'J':
			P = "long";
			break;
		case 'S':
			P = "short";
			break;
		case 'Z':
			P = "boolean";
			break;
		case 'V':
			P = "void";
			break;
		case ';':
			continue;
		default:
			assert(0);
			return;
		}
		if (parameter > 1) {
			fputs(", ", stdoutF);
		}
		fputs(P, stdoutF);
show_array:
		for (; dimension > 0; --dimension) {
			fputs("[]", stdoutF);
		}
	}	
}

void
Cfunction::dump_descriptor(void) const
{
	methodT	*methodP;
	stringT	**parametersPP;
	u4T		parameters, i;

	dump_access(m_access_flags);
	if ((methodP = m_methodP)) {
		protoT	*protoP = methodP->protoP;
		if (protoP) {
			if (protoP->return_typeP) {
				dump_descriptor(protoP->return_typeP->P);
				fputc(' ', stdoutF);
		}	}
		if (methodP->classP) {
			dump_descriptor(methodP->classP->P);
			fputs("::", stdoutF);
		}
		if (methodP->nameP) {
			dumpUnicode(stdoutF, methodP->nameP->P);
		}
		fputc('(', stdoutF);
		if (protoP && (parametersPP = protoP->parametersPP) && (parameters = protoP->parameters)) {
			for (i = 0; ; ) {
				dump_descriptor(parametersPP[i]->P);
				if (++i == parameters) {
					break;
				}
				fputs(", ", stdoutF);
		}	}
		fputc(')', stdoutF);
}	}

extern FILE *stdoutF;

static void
print_register(registerT *registerP) 
{
	variableT	*variableP;

	fprintf(stdoutF, "v%d", registerP->number);
	if ((variableP = registerP->u.variableP)) {
		fprintf(stdoutF, "/* ");
		dumpUnicode(stdoutF, variableP->nameP->P);
		fprintf(stdoutF, " */");
	}
}

static void
print_string(registerT *registerP) 
{
	fputc('"', stdoutF);
	dumpUnicode(stdoutF, registerP->u.stringP->P);
	fputc('"', stdoutF);
}

static void
print_type(registerT *registerP) 
{
	Cfunction::dump_descriptor(registerP->u.typeP->P);
}

static void
print_field(registerT *registerP) 
{
	dumpUnicode(stdoutF, registerP->u.fieldP->nameP->P);
}

static void
print_method(registerT *registerP) 
{
	dumpUnicode(stdoutF, registerP->u.methodP->nameP->P);
}

static void
print_branch(codeT *assemblerP) 
{
	codeT	*targetP = assemblerP->m_jump.P;
	
	if (!targetP) {
		fprintf(stdoutF, "+?");
	} else {
		fprintf(stdoutF, "%+d", targetP->m_pc - assemblerP->m_pc);
}	}

void
Cfunction::dump_pcode(codeT *assemblerP)
{
	int				code;
	const char		*nameP;
	positionT		*positionP;

	positionP = assemblerP->m_positionP;
	if (positionP) {
		fprintf(stdoutF, "%u", positionP->line);
	}
	fprintf(stdoutF, "[%u] ", (u4T) assemblerP->m_pc);
	code       = assemblerP->m_pcode;
	nameP      = opcodes[code];
	fprintf(stdoutF, "%s", nameP);
	switch (code) {
	case NOP:				// 0x00: 10x ->  00|op  -> op
	case RETURN_VOID:		// 0x0e:
		break;
	case MOVE_RESULT:		// 0x0a: 11x ->  AA|op  -> op vAA 
	case MOVE_RESULT_WIDE:	// 0x0b
	case MOVE_RESULT_OBJECT:// 0x0c
	case MOVE_EXCEPTION:	// 0x0d
	case RETURN:			// 0x0f
	case RETURN_WIDE:		// 0x10
	case RETURN_OBJECT:		// 0x11	
	case MONITOR_ENTER:		// 0x1d
	case MONITOR_EXIT:		// 0x1e
	case THROW:				// 0x27
		fputc('\t', stdoutF);
		print_register(&(assemblerP->info.register1P->reg1));
		break;
	case ARRAY_LENGTH:		// 0x21
	case NEG_INT:			// 0x7b:
	case NOT_INT:			// 0x7c:
	case NEG_LONG:			// 0x7d:
	case NOT_LONG:			// 0x7e:
	case NEG_FLOAT:			// 0x7f:
	case NEG_DOUBLE:		// 0x80:
	case INT_TO_LONG:		// 0x81:
	case INT_TO_FLOAT:		// 0x82:
	case INT_TO_DOUBLE:		// 0x83:
	case LONG_TO_INT:		// 0x84:
	case LONG_TO_FLOAT:		// 0x85:
	case LONG_TO_DOUBLE:	// 0x86:
	case FLOAT_TO_INT:		// 0x87:
	case FLOAT_TO_LONG:		// 0x88:
	case FLOAT_TO_DOUBLE:	// 0x89:
	case DOUBLE_TO_INT:		// 0x8a:
	case DOUBLE_TO_LONG:	// 0x8b:
	case DOUBLE_TO_FLOAT:	// 0x8c:
	case INT_TO_BYTE:		// 0x8d:
	case INT_TO_CHAR:		// 0x8e:
	case INT_TO_SHORT:		// 0x8f:
	case ADD_INT2:			// 0xb0:
	case SUB_INT2:			// 0xb1:
	case MUL_INT2:			// 0xb2:
	case DIV_INT2:			// 0xb3:
	case REM_INT2:			// 0xb4:
	case AND_INT2:			// 0xb5:
	case OR_INT2:			// 0xb6:
	case XOR_INT2:			// 0xb7:
	case SHIFT_LEFT_INT2:	// 0xb8:
	case SHIFT_RIGHT_INT2:	// 0xb9:
	case USHIFT_RIGHT_INT2:	// 0xba:
	case ADD_LONG2:			// 0xbb:
	case SUB_LONG2:			// 0xbc:
	case MUL_LONG2:			// 0xbd:
	case DIV_LONG2:			// 0xbe:
	case REM_LONG2:			// 0xbf:
	case AND_LONG2:			// 0xc0:
	case OR_LONG2:			// 0xc1:
	case XOR_LONG2:			// 0xc2:
	case SHIFT_LEFT_LONG2:	// 0xc3:
	case SHIFT_RIGHT_LONG2:	// 0xc4:
	case USHIFT_RIGHT_LONG2:// 0xc5:
	case ADD_FLOAT2:		// 0xc6:
	case SUB_FLOAT2:		// 0xc7:
	case MUL_FLOAT2:		// 0xc8:
	case DIV_FLOAT2:		// 0xc9:
	case REM_FLOAT2:		// 0xca:
	case ADD_DOUBLE2:		// 0xcb:
	case SUB_DOUBLE2:		// 0xcc:
	case MUL_DOUBLE2:		// 0xcd:
	case DIV_DOUBLE2:		// 0xce:
	case REM_DOUBLE2:		// 0xcf:

	case MOVE_4_4:			// 0x01: 12x -> B|A|op -> op vA, vB
	case MOVE_WIDE_4_4:		// 0x04
	case MOVE_OBJECT_4_4:	// 0x07
	case MOVE_8_16:			// 0x02: 22x -> AA|op BBBB -> op vAA, vBBBB
	case MOVE_WIDE_8_16:	// 0x05:
	case MOVE_OBJECT_8_16:	// 0x08:
	case MOVE_16_16:		// 0x03: 32x -> |op AAAA BBBB -> op vAAAA, vBBBB
	case MOVE_WIDE_16_16:	// 0x06:
	case MOVE_OBJECT_16_16:	// 0x09:
	{
		info_register2T	*infoP = assemblerP->info.register2P;

		fputc('\t', stdoutF);
		print_register(&(infoP->reg1));
		fputs(", ", stdoutF);
		print_register(&(infoP->reg2));
		break;
	}
	case CMPL_FLOAT:		// 0x2d: 23x -> AA|op CC|BB -> op vAA, vBB, vCC
	case CMPG_FLOAT:		// 0x2e:
	case CMPL_DOUBLE:		// 0x2f:
	case CMPG_DOUBLE:		// 0x30:
	case CMP_LONG:			// 0x31:

	case AGET:				// 0x44:
	case AGET_WIDE:			// 0x45:
	case AGET_OBJECT:		// 0x46:
	case AGET_BOOLEAN:		// 0x47:
	case AGET_BYTE:			// 0x48:
	case AGET_CHAR:			// 0x49:
	case AGET_SHORT:		// 0x4a:
	case APUT:				// 0x4b:
	case APUT_WIDE:			// 0x4c:
	case APUT_OBJECT:		// 0x4d:
	case APUT_BOOLEAN:		// 0x4e:
	case APUT_BYTE:			// 0x4f:
	case APUT_CHAR:			// 0x50:
	case APUT_SHORT:		// 0x51:

	case ADD_INT:			// 0x90:
	case SUB_INT:			// 0x91:
	case MUL_INT:			// 0x92:
	case DIV_INT:			// 0x93:
	case REM_INT:			// 0x94:
	case AND_INT:			// 0x95:
	case OR_INT:			// 0x96:
	case XOR_INT:			// 0x97:
	case SHIFT_LEFT_INT:	// 0x98:
	case SHIFT_RIGHT_INT:	// 0x99:
	case USHIFT_RIGHT_INT:	// 0x9a:
	case ADD_LONG:			// 0x9b:
	case SUB_LONG:			// 0x9c:
	case MUL_LONG:			// 0x9d:
	case DIV_LONG:			// 0x9e:
	case REM_LONG:			// 0x9f:
	case AND_LONG:			// 0xa0:
	case OR_LONG:			// 0xa1:
	case XOR_LONG:			// 0xa2:
	case SHIFT_LEFT_LONG:	// 0xa3:
	case SHIFT_RIGHT_LONG:	// 0xa4:
	case USHIFT_RIGHT_LONG:	// 0xa5:
	case ADD_FLOAT:			// 0xa6:
	case SUB_FLOAT:			// 0xa7:
	case MUL_FLOAT:			// 0xa8:
	case DIV_FLOAT:			// 0xa9:
	case REM_FLOAT:			// 0xaa:
	case ADD_DOUBLE:		// 0xab:
	case SUB_DOUBLE:		// 0xac:
	case MUL_DOUBLE:		// 0xad:
	case DIV_DOUBLE:		// 0xae:
	case REM_DOUBLE:		// 0xaf:
	{
		info_register3T	*infoP = assemblerP->info.register3P;

		fputc('\t', stdoutF);
		print_register(&(infoP->reg1));
		fputs(", ", stdoutF);
		print_register(&(infoP->reg2));
		fputs(", ", stdoutF);
		print_register(&(infoP->reg3));
		break;
	}
	case CONST_4_4:			// 0x12: 11n -> B|A|op -> op vA, #+B 
	case CONST_8_16:		// 0x13: 21s -> AA|op BBBB -> op vAA, #+BBBB
	case CONST_8_32:		// 0x14: 31i -> AA|op BBBBlo BBBBhi -> op vAA, #+BBBBBBBB
	{
		info_valueT	*infoP = assemblerP->info.valueP;

		fputc('\t', stdoutF);
		print_register(&(infoP->reg1));
		fprintf(stdoutF, ", #%+d", infoP->value);
		break;
	}
	case CONST_HIGH:		// 0x15: 21h -> AA|op BBBB -> op vAA, #+BBBB0000
	{
		info_valueT	*infoP = assemblerP->info.valueP;

		fputc('\t', stdoutF);
		print_register(&(infoP->reg1));
		fprintf(stdoutF, ", #+%04x0000", infoP->value);
		break;
	}		
	case CONST_WIDE_HIGH:	// 0x19: 21h -> AA|op BBBB -> op vAA, #+BBBB000000000000 
	{
		info_valueT	*infoP = assemblerP->info.valueP;

		fputc('\t', stdoutF);
		print_register(&(infoP->reg1));
		fprintf(stdoutF, ", #+%04x000000000000", infoP->value);
		break;
	}		
	case CONST_WIDE_8_16:	// 0x16: 21s -> const-wide/16 vAA, #+BBBB
	case CONST_WIDE_8_32:	// 0x17:
	case CONST_WIDE_8_64:	// 0x18: 51l -> AA|op BBBBlo BBBB BBBB BBBBhi -> op vAA, #+BBBBBBBBBBBBBBBB
	{
		info_wideT	*infoP = assemblerP->info.wideP;

		fputc('\t', stdoutF);
		print_register(&(infoP->reg1));
#ifdef WIN32
		fprintf(stdoutF, ", #%+I64d", infoP->value);
#else
		fprintf(stdoutF, ", #%+lld", infoP->value);
#endif
		break;
	}
	case ADD_INT_LIT8:		// 0xd8: 22b -> AA|op CC|BB -> op vAA, vBB, #+CC
	case REVERSE_SUB_INT_LIT8: // 0xd9:
	case MUL_INT_LIT8:		// 0xda:
	case DIV_INT_LIT8:		// 0xdb:
	case REM_INT_LIT8:		// 0xdc:
	case AND_INT_LIT8:		// 0xdd:
	case OR_INT_LIT8:		// 0xde:
	case XOR_INT_LIT8:		// 0xdf:
	case SHIFT_LEFT_INT_LIT8://0xe0:
	case SHIFT_RIGHT_INT_LIT8://0xe1:
	case USHIFT_RIGHT_INT_LIT8://0xe2:
	case ADD_INT_LIT16:		// 0xd0: 22s -> B|A|op CCCC -> op vA, vB, #+CCCC
	case REVERSE_SUB_INT_LIT16:	// 0xd1:
	case MUL_INT_LIT16:		// 0xd2:
	case DIV_INT_LIT16:		// 0xd3:
	case REM_INT_LIT16:		// 0xd4:
	case AND_INT_LIT16:		// 0xd5:
	case OR_INT_LIT16:		// 0xd6:
	case XOR_INT_LIT16:		// 0xd7:	
	{
		info_value2T	*infoP = assemblerP->info.value2P;

		fputc('\t', stdoutF);
		print_register(&(infoP->reg1));
		fputs(", ", stdoutF);
		print_register(&(infoP->reg2));
		fprintf(stdoutF, ", #%+d", infoP->value);
		break;
	}
	case CONST_STRING_8_16: // 0x1a: 21c -> AA|op BBBB -> op vAA, string@BBBB
	case CONST_STRING_8_32:	// 0x1b: 31c -> AA|op BBBBlo BBBBhi -> op vAA, string@BBBBBBBB
	{
		info_register2T	*infoP = assemblerP->info.register2P;

		fputc('\t', stdoutF);
		print_register(&(infoP->reg1));
		fputs(", ", stdoutF);
		print_string(&(infoP->reg2));
		break;
	}
	case CONST_CLASS:		// 0x1c: 21c -> AA|op BBBB -> op vAA, type@BBBB
	case CHECK_CAST:		// 0x1f:
	case NEW_INSTANCE:		// 0x22:
	{
		info_register2T	*infoP = assemblerP->info.register2P;

		fputc('\t', stdoutF);
		print_register(&(infoP->reg1));
		fputs(", ", stdoutF);
		print_type(&(infoP->reg2));
		break;
	}
	case STATIC_GET:		// 0x60: 21c -> AA|op BBBB -> op vAA, field@BBBB
	case STATIC_GET_WIDE:	// 0x61:
	case STATIC_GET_OBJECT:	// 0x62:
	case STATIC_GET_BOOLEAN:// 0x63:
	case STATIC_GET_BYTE:	// 0x64:
	case STATIC_GET_CHAR:	// 0x65:
	case STATIC_GET_SHORT:	// 0x66:
	case STATIC_PUT:		// 0x67:
	case STATIC_PUT_WIDE:	// 0x68:
	case STATIC_PUT_OBJECT:	// 0x69:
	case STATIC_PUT_BOOLEAN:// 0x6a:
	case STATIC_PUT_BYTE:	// 0x6b:
	case STATIC_PUT_CHAR:	// 0x6c:
	case STATIC_PUT_SHORT:	// 0x6d:
	{
		info_register2T	*infoP = assemblerP->info.register2P;

		fputc('\t', stdoutF);
		print_register(&(infoP->reg1));
		fputs(", ", stdoutF);
		print_field(&(infoP->reg2));
		break;
	}
	case INSTANCE_OF:		// 0x20: 22c -> B|A|op CCCC -> op vA, vB, type@CCCC
	case NEW_ARRAY:			// 0x23:
	{
		info_register3T	*infoP = assemblerP->info.register3P;

		fputc('\t', stdoutF);
		print_register(&(infoP->reg1));
		fputs(", ", stdoutF);
		print_register(&(infoP->reg2));
		fputs(", ", stdoutF);
		print_type(&(infoP->reg3));
		break;
	}
	case IGET:				// 0x52: 22c -> B|A|op CCCC -> op vA, vB, field@CCCC
	case IGET_WIDE:			// 0x53:
	case IGET_OBJECT:		// 0x54:
	case IGET_BOOLEAN:		// 0x55:
	case IGET_BYTE:			// 0x56:
	case IGET_CHAR:			// 0x57:
	case IGET_SHORT:		// 0x58:
	case IPUT:				// 0x59:
	case IPUT_WIDE:			// 0x5a:
	case IPUT_OBJECT:		// 0x5b:
	case IPUT_BOOLEAN:		// 0x5c:
	case IPUT_BYTE:			// 0x5d:
	case IPUT_CHAR:			// 0x5e:
	case IPUT_SHORT:		// 0x5f:
	{
		info_register3T	*infoP = assemblerP->info.register3P;

		fputc('\t', stdoutF);
		print_register(&(infoP->reg1));
		fputs(", ", stdoutF);
		print_register(&(infoP->reg2));
		fputs(", ", stdoutF);
		print_field(&(infoP->reg3));
		break;
	}
	case GOTO_8:			// 0x28: 10t ->  AA|op -> op +AA
	case GOTO_16:			// 0x29: 20t -> |op AAAA -> op +AAAA
	case GOTO_32:			// 0x2a:		// 30t -> |op AAAAlo AAAAhi -> op +AAAAAAAA
	{
		fputc('\t', stdoutF);
		print_branch(assemblerP);
		break;
	}
	case IF_EQ_ZERO:		// 0x38: 21t -> AA|op BBBB -> op vAA, +BBBB
	case IF_NE_ZERO:		// 0x39:
	case IF_LT_ZERO:		// 0x3a:
	case IF_GE_ZERO:		// 0x3b:
	case IF_GT_ZERO:		// 0x3c:
	case IF_LE_ZERO:		// 0x3d:
	{
		info_register1T	*infoP = assemblerP->info.register1P;

		fputc('\t', stdoutF);
		print_register(&(infoP->reg1));
		fputs(", ", stdoutF);
		print_branch(assemblerP);
		break;
	}
	case IF_EQ:				// 0x32: 22t -> B|A|op CCCC -> op vA, vB, +CCCC
	case IF_NE:				// 0x33:
	case IF_LT:				// 0x34:
	case IF_GE:				// 0x35:
	case IF_GT:				// 0x36:
	case IF_LE:				// 0x37:
	{
		info_register2T	*infoP = assemblerP->info.register2P;

		fputc('\t', stdoutF);
		print_register(&(infoP->reg1));
		fputs(", ", stdoutF);
		print_register(&(infoP->reg2));
		fputs(", ", stdoutF);
		print_branch(assemblerP);
		break;
	}
	case FILL_ARRAY_DATA:	// 0x26: 31t -> AA|op BBBBlo BBBBhi -> op vAA, +BBBBBBBB fill-array-data vAA, +BBBBBBBB (with supplemental data as specified below in "fill-array-data Format")
	{
		info_fillT	*infoP = assemblerP->info.fillP;
		u4T	width          = infoP->width;
		u4T	entries        = infoP->entries;
		u4T data           = width * entries;
		u4T	i, j;

		if (width == 1) {
			fputc('\t', stdoutF);
			print_register(&(infoP->reg1));
			fprintf(stdoutF, ", array[%d]={", entries);
			for (i = 0; ; ) {
				fprintf(stdoutF, "%2x", infoP->data[i]);
				if (++i == data) {
					break;
				}
				fputs(", ", stdoutF);
			}
		} else {
			fputc('\t', stdoutF);
			print_register(&(infoP->reg1));
			fprintf(stdoutF, ", array[%d][%d]={", entries, width);
			assert(width > 1);
			j = 0;
			for (i = 0; ; ) {
				if (!j) {
					fprintf(stdoutF, "{");
					j = width;
				}
				fprintf(stdoutF,  "%2x", infoP->data[i]);
				if (!--j) {
					fputc('}', stdoutF);
				}
				if (++i == data) {
					break;
				}
				fputs(", ", stdoutF);
		}	}
		fputc('}', stdoutF);
		break;
	}
	case PACKED_SWITCH:		// 0x2b: 31t -> AA|op BBBBlo BBBBhi -> op vAA, +BBBBBBBB (with supplemental data as specified below in "packed-switch Format")
	case SPARSE_SWITCH:		// 0x2c: 31t -> AA|op BBBBlo BBBBhi -> op vAA, +BBBBBBBB  (with supplemental data as specified below in "sparse-switch Format")
	{
		info_valueT *infoP = assemblerP->info.valueP;

		fputc('\t', stdoutF);
		print_register(&(infoP->reg1));
		fprintf(stdoutF, ", %d, ", infoP->value);
		print_branch(assemblerP);
		break;
	}
	case FILL_ARRAY:		// 0x24: 35c -> B|A|op CCCC G|F|E|D -> [B=5] op {vD, vE, vF, vG, vA}, type@CCCC
	case FILLED_NEW_ARRAY:	// 0x25: 3rc -> AA|op BBBB CCCC -> op {vCCCC .. vNNNN}, type@BBBB
	{
		info_registersT	*infoP = assemblerP->info.registersP;
		u4T				size   = infoP->size;
		u4T				i;

		fputs("\t{", stdoutF);

		if (size) {
			for (i = 0; ; ) {
				print_register(infoP->reg + i);
				if (++i == size) {
					break;
				}
				fputs(", ", stdoutF);
		}	}
		fputs("}, ", stdoutF);
		print_type(&(infoP->last));
		break;
	}
	case INVOKE_VIRTUAL_RANGE: // 0x74: 3rc -> AA|op BBBB CCCC -> op {vCCCC .. vNNNN}, meth@BBBB
	case INVOKE_SUPER_RANGE://0x75:
	case INVOKE_DIRECT_RANGE://0x76:
	case INVOKE_STATIC_RANGE://0x77:
	case INVOKE_INTERFACE_RANGE://0x78:
	case INVOKE_VIRTUAL:	// 0x6e: [B=5] op {vD, vE, vF, vG, vA}, meth@CCCC
	case INVOKE_SUPER:		// 0x6f: [B=4] op {vD, vE, vF, vG}, meth@CCCC
	case INVOKE_DIRECT:		// 0x70: [B=3] op {vD, vE, vF}, meth@CCCC
	case INVOKE_STATIC:		// 0x71: [B=2] op {vD, vE}, meth@CCCC
	case INVOKE_INTERFACE:	// 0x72: [B=1] op {vD}, meth@CCCC	
	{
		info_registersT	*infoP = assemblerP->info.registersP;
		u4T				size   = infoP->size;
		u4T				i;

		fputs("\t{", stdoutF);

		if (size) {
			for (i = 0; ; ) {
				print_register(infoP->reg + i);
				if (++i == size) {
					break;
				}
				fputs(", ", stdoutF);
		}	}
		fputs("}, ", stdoutF);
		print_method(&(infoP->last));
		break;
	}
	case TRY_START:
	{
		info_tryT	*infoP = assemblerP->info.tryP;

		fputc('\t', stdoutF);
		if (infoP->classtype.u.typeP) {
			print_type(&(infoP->classtype));
		}
		print_branch(assemblerP);
		break;
	}
	case TRY_END:
	{
		info_tryT	*infoP = assemblerP->info.tryP;

		if (infoP->classtype.u.typeP) {
			fputc('\t', stdoutF);
			print_type(&(infoP->classtype));
		}
		break;
	}
	default:
		assert(0);
	}
}

void
Cfunction::emit_ta(void)
{
	Cclass		*classP;
	positionT	*positionP;
	Cfile		*fileP;

	classP = m_classP;
	if (!classP->m_has_clones) {
		classP->m_has_clones = 1;
		classP->emit_ta();
	}

	fprintf(taF, "$INSTANCE F%p F\n", this);
	fprintf(taF, "contain S%p F%p\n", classP, this);
	fprintf(tempF, "F%p {label=\"", this);
	dumpUnicode(tempF, m_methodP->nameP->P);
	fputc('"', tempF);
	fileP     = classP->m_source_fileP;
	positionP = m_assemblersP->m_positionP;
	if (positionP) {
		fprintf(tempF, " lineno=%d", positionP->line);
		if (positionP->source_fileP) {
			fileP = positionP->source_fileP;
	}	}
	if (fileP == classP->m_source_fileP) {
		fprintf(tempF, " file=\".\"");
	} else {
		fprintf(tempF, " file=\"");
		dumpUnicode(tempF, fileP->m_nameP->P);
		fputc('"', tempF);
	}
	fputs("}\n", tempF);
}

Cclone *
Cfunction::locate_clone(const codeT *startP, const codeT *endP)
{
	Cclone		*cloneP;
	int			start, lth, lth_code;
	positionT	*positionP;
	Cfile		*fileP;

	for (cloneP = m_clonesP; cloneP; cloneP = cloneP->m_nextP) {
		if (cloneP->m_startP == startP && cloneP->m_endP == endP) {
			return (cloneP);
	}	}
	if (!m_clonesP) {
		emit_ta();
	}
	cloneP = new Cclone(startP, endP);
	cloneP->m_nextP = m_clonesP;
	m_clonesP       = cloneP;

	fprintf(taF, "$INSTANCE X%p X\n", cloneP);
	fprintf(taF, "contain F%p X%p\n", this, cloneP);

	start    = startP - m_assemblersP;
	lth      = endP   - startP;
	lth_code = m_assemblers - start;
	fprintf(tempF, "X%p {label=\"%d+%d\" xrel=%d yrel=%d",
		cloneP,
		start, lth,
		relativeToShort(((double) start)/((double) lth_code)),
		relativeToShort(((double) lth)/((double) lth_code))
	);
	if ((positionP = startP->m_positionP)) {
		fileP = positionP->source_fileP;
		if (!fileP) {
			fileP = m_classP->m_source_fileP;
		}
		if (fileP) {
			if (fileP == m_classP->m_source_fileP) {
				fputs(" file=\".\"", tempF);
			} else {
				fputs(" file=\"", tempF);
				dumpUnicode(tempF, fileP->m_nameP->P);
				fputc('"', tempF);
			}
			fprintf(tempF, "lineno=%d", positionP->line);
	}	}
	fputs("}\n", tempF);

	return(cloneP);
}
