#include "typedef.h"

#include <assert.h>

#include "object.h"
#include "opcode.h"
#include "code.h"

#define HASH_SIZE 0x200000

static codeT	**hash_table[HASH_SIZE] = {0};
static u4T		g_hash;

static void
hash_integer(u4T val)
{
	g_hash *= 0x811C9DC5;	// Large prime
    g_hash ^= val;
}

static void
hash_long(u8T val)
{
	hash_integer((u4T) val);
	hash_integer((u4T) (val >> 32));
}
static void
hash_data(u4T lth, u1T *dataP)
{
	u1T *P;
	u1T	*endP;

	P = dataP;
	for (endP = P + lth; P < endP; ++P) {
		hash_integer(*P);
}	}

static void
hash_string(stringT *stringP)
{
	hash_data((stringP->lth<<1), (u1T *) stringP->P);
}

static void
hash_type(stringT *stringP)
{
	hash_string(stringP);
}

static void
hash_field(fieldT *fieldP)
{
	hash_string(fieldP->nameP);
}

static void
hash_method(methodT *methodP)
{
	hash_string(methodP->nameP);
}

static void
hash_register(registerT *registerP)
{
	variableT	*variableP;

	if ((variableP = registerP->u.variableP)) {
		hash_string(variableP->nameP);
	}
	hash_integer(registerP->number);
}

void
hash_pcode(codeT *assemblerP)
{
	int		 pcode;
	codeT	 **hashPP;

	g_hash = pcode = assemblerP->m_pcode;

	switch (pcode) {
	case NOP:				// 0x00: 10x ->  00|op  -> op
	case RETURN_VOID:		// 0x0e:
		break;
	case MOVE_RESULT:		// 0x0a: 11x ->  AA|op  -> op vAA 
	case MOVE_RESULT_WIDE:	// 0x0b
	case MOVE_RESULT_OBJECT:// 0x0c
	case MOVE_EXCEPTION:	// 0x0d
	case RETURN:			// 0x0f
	case RETURN_WIDE:		// 0x10
	case RETURN_OBJECT:		// 0x11
	case MONITOR_ENTER:		// 0x1d
	case MONITOR_EXIT:		// 0x1e
	case THROW:				// 0x27
		hash_register(&(assemblerP->info.register1P->reg1));
		break;
	case ARRAY_LENGTH:		// 0x21: 12x -> B|A|op -> op vA, vB 
	case NEG_INT:			// 0x7b:
	case NOT_INT:			// 0x7c:
	case NEG_LONG:			// 0x7d:
	case NOT_LONG:			// 0x7e:
	case NEG_FLOAT:			// 0x7f:
	case NEG_DOUBLE:		// 0x80:
	case INT_TO_LONG:		// 0x81:
	case INT_TO_FLOAT:		// 0x82:
	case INT_TO_DOUBLE:		// 0x83:
	case LONG_TO_INT:		// 0x84:
	case LONG_TO_FLOAT:		// 0x85:
	case LONG_TO_DOUBLE:	// 0x86:
	case FLOAT_TO_INT:		// 0x87:
	case FLOAT_TO_LONG:		// 0x88:
	case FLOAT_TO_DOUBLE:	// 0x89:
	case DOUBLE_TO_INT:		// 0x8a:
	case DOUBLE_TO_LONG:	// 0x8b:
	case DOUBLE_TO_FLOAT:	// 0x8c:
	case INT_TO_BYTE:		// 0x8d:
	case INT_TO_CHAR:		// 0x8e:
	case INT_TO_SHORT:		// 0x8f:
	case ADD_INT2:			// 0xb0:
	case SUB_INT2:			// 0xb1:
	case MUL_INT2:			// 0xb2:
	case DIV_INT2:			// 0xb3:
	case REM_INT2:			// 0xb4:
	case AND_INT2:			// 0xb5:
	case OR_INT2:			// 0xb6:
	case XOR_INT2:			// 0xb7:
	case SHIFT_LEFT_INT2:	// 0xb8:
	case SHIFT_RIGHT_INT2:	// 0xb9:
	case USHIFT_RIGHT_INT2:	// 0xba:
	case ADD_LONG2:			// 0xbb:
	case SUB_LONG2:			// 0xbc:
	case MUL_LONG2:			// 0xbd:
	case DIV_LONG2:			// 0xbe:
	case REM_LONG2:			// 0xbf:
	case AND_LONG2:			// 0xc0:
	case OR_LONG2:			// 0xc1:
	case XOR_LONG2:			// 0xc2:
	case SHIFT_LEFT_LONG2:	// 0xc3:
	case SHIFT_RIGHT_LONG2:	// 0xc4:
	case USHIFT_RIGHT_LONG2:// 0xc5:
	case ADD_FLOAT2:		// 0xc6:
	case SUB_FLOAT2:		// 0xc7:
	case MUL_FLOAT2:		// 0xc8:
	case DIV_FLOAT2:		// 0xc9:
	case REM_FLOAT2:		// 0xca:
	case ADD_DOUBLE2:		// 0xcb:
	case SUB_DOUBLE2:		// 0xcc:
	case MUL_DOUBLE2:		// 0xcd:
	case DIV_DOUBLE2:		// 0xce:
	case REM_DOUBLE2:		// 0xcf:
	case MOVE_4_4:			// 0x01: 12x -> B|A|op -> op vA, vB
	case MOVE_WIDE_4_4:		// 0x04
	case MOVE_OBJECT_4_4:	// 0x07
	case MOVE_8_16:			// 0x02: 22x -> AA|op BBBB -> op vAA, vBBBB
	case MOVE_WIDE_8_16:	// 0x05:
	case MOVE_OBJECT_8_16:	// 0x08:
	case MOVE_16_16:		// 0x03: 32x -> |op AAAA BBBB -> op vAAAA, vBBBB
	case MOVE_WIDE_16_16:	// 0x06:
	case MOVE_OBJECT_16_16:	// 0x09:
	{
		info_register2T	*registersP;

		registersP = assemblerP->info.register2P;
		hash_register(&(registersP->reg1));
		hash_register(&(registersP->reg2));
		break;
	}
	case CMPL_FLOAT:		// 0x2d: 23x -> AA|op CC|BB -> op vAA, vBB, vCC
	case CMPG_FLOAT:		// 0x2e:
	case CMPL_DOUBLE:		// 0x2f:
	case CMPG_DOUBLE:		// 0x30:
	case CMP_LONG:			// 0x31:

	case AGET:				// 0x44:
	case AGET_WIDE:			// 0x45:
	case AGET_OBJECT:		// 0x46:
	case AGET_BOOLEAN:		// 0x47:
	case AGET_BYTE:			// 0x48:
	case AGET_CHAR:			// 0x49:
	case AGET_SHORT:		// 0x4a:
	case APUT:				// 0x4b:
	case APUT_WIDE:			// 0x4c:
	case APUT_OBJECT:		// 0x4d:
	case APUT_BOOLEAN:		// 0x4e:
	case APUT_BYTE:			// 0x4f:
	case APUT_CHAR:			// 0x50:
	case APUT_SHORT:		// 0x51:

	case ADD_INT:			// 0x90:
	case SUB_INT:			// 0x91:
	case MUL_INT:			// 0x92:
	case DIV_INT:			// 0x93:
	case REM_INT:			// 0x94:
	case AND_INT:			// 0x95:
	case OR_INT:			// 0x96:
	case XOR_INT:			// 0x97:
	case SHIFT_LEFT_INT:	// 0x98:
	case SHIFT_RIGHT_INT:	// 0x99:
	case USHIFT_RIGHT_INT:	// 0x9a:
	case ADD_LONG:			// 0x9b:
	case SUB_LONG:			// 0x9c:
	case MUL_LONG:			// 0x9d:
	case DIV_LONG:			// 0x9e:
	case REM_LONG:			// 0x9f:
	case AND_LONG:			// 0xa0:
	case OR_LONG:			// 0xa1:
	case XOR_LONG:			// 0xa2:
	case SHIFT_LEFT_LONG:	// 0xa3:
	case SHIFT_RIGHT_LONG:	// 0xa4:
	case USHIFT_RIGHT_LONG:	// 0xa5:
	case ADD_FLOAT:			// 0xa6:
	case SUB_FLOAT:			// 0xa7:
	case MUL_FLOAT:			// 0xa8:
	case DIV_FLOAT:			// 0xa9:
	case REM_FLOAT:			// 0xaa:
	case ADD_DOUBLE:		// 0xab:
	case SUB_DOUBLE:		// 0xac:
	case MUL_DOUBLE:		// 0xad:
	case DIV_DOUBLE:		// 0xae:
	case REM_DOUBLE:		// 0xaf:
	{
		info_register3T	*registersP;

		registersP = assemblerP->info.register3P;
		hash_register(&(registersP->reg1));
		hash_register(&(registersP->reg2));
		hash_register(&(registersP->reg3));
		break;
	}
	case CONST_4_4:			// 0x12: 11n -> B|A|op -> op vA, #+B 
	case CONST_8_16:		// 0x13: 21s -> AA|op BBBB -> op vAA, #+BBBB
	case CONST_8_32:		// 0x14: 31i -> AA|op BBBBlo BBBBhi -> op vAA, #+BBBBBBBB
	case CONST_HIGH:		// 0x15: 21h -> AA|op BBBB -> op vAA, #+BBBB0000
	case CONST_WIDE_HIGH:	// 0x19: 21h -> AA|op BBBB -> op vAA, #+BBBB000000000000 
	{
		info_valueT	*registersP;

		registersP = assemblerP->info.valueP;
		hash_register(&(registersP->reg1));
		hash_integer(registersP->value);
		break;
	}
	case CONST_WIDE_8_16:	// 0x16:
	case CONST_WIDE_8_32:	// 0x17:
	case CONST_WIDE_8_64:	// 0x18: 51l -> AA|op BBBBlo BBBB BBBB BBBBhi -> op vAA, #+BBBBBBBBBBBBBBBB
	{
		info_wideT	*registersP;

		registersP = assemblerP->info.wideP;
		hash_register(&(registersP->reg1));
		hash_long(registersP->value);
		break;
	}
	case ADD_INT_LIT8:		// 0xd8: 22b -> AA|op CC|BB -> op vAA, vBB, #+CC
	case REVERSE_SUB_INT_LIT8: // 0xd9:
	case MUL_INT_LIT8:		// 0xda:
	case DIV_INT_LIT8:		// 0xdb:
	case REM_INT_LIT8:		// 0xdc:
	case AND_INT_LIT8:		// 0xdd:
	case OR_INT_LIT8:		// 0xde:
	case XOR_INT_LIT8:		// 0xdf:
	case SHIFT_LEFT_INT_LIT8://0xe0:
	case SHIFT_RIGHT_INT_LIT8://0xe1:
	case USHIFT_RIGHT_INT_LIT8://0xe2:
	case ADD_INT_LIT16:		// 0xd0: 22s -> B|A|op CCCC -> op vA, vB, #+CCCC
	case REVERSE_SUB_INT_LIT16:	// 0xd1:
	case MUL_INT_LIT16:		// 0xd2:
	case DIV_INT_LIT16:		// 0xd3:
	case REM_INT_LIT16:		// 0xd4:
	case AND_INT_LIT16:		// 0xd5:
	case OR_INT_LIT16:		// 0xd6:
	case XOR_INT_LIT16:		// 0xd7:
	{
		info_value2T	*registersP;

		registersP = assemblerP->info.value2P;
		hash_register(&(registersP->reg1));
		hash_register(&(registersP->reg2));
		hash_integer(registersP->value);
		break;
	}
	case CONST_STRING_8_16: // 0x1a: 21c -> AA|op BBBB -> op vAA, string@BBBB
	case CONST_STRING_8_32:	// 0x1b: 31c -> AA|op BBBBlo BBBBhi -> op vAA, string@BBBBBBBB
	{
		info_register2T	*registersP;

		registersP = assemblerP->info.register2P;
		hash_register(&(registersP->reg1));
		hash_string(registersP->reg2.u.stringP);
		break;
	}
	case CONST_CLASS:		// 0x1c: 21c -> AA|op BBBB -> op vAA, type@BBBB
	case CHECK_CAST:		// 0x1f:
	case NEW_INSTANCE:		// 0x22:
	{
		info_register2T	*registersP;

		registersP = assemblerP->info.register2P;
		hash_register(&(registersP->reg1));
		hash_type(registersP->reg2.u.typeP);
		break;
	}
	case STATIC_GET:		// 0x60: 21c -> AA|op BBBB -> op vAA, field@BBBB
	case STATIC_GET_WIDE:	// 0x61:
	case STATIC_GET_OBJECT:	// 0x62:
	case STATIC_GET_BOOLEAN:// 0x63:
	case STATIC_GET_BYTE:	// 0x64:
	case STATIC_GET_CHAR:	// 0x65:
	case STATIC_GET_SHORT:	// 0x66:
	case STATIC_PUT:		// 0x67:
	case STATIC_PUT_WIDE:	// 0x68:
	case STATIC_PUT_OBJECT:	// 0x69:
	case STATIC_PUT_BOOLEAN:// 0x6a:
	case STATIC_PUT_BYTE:	// 0x6b:
	case STATIC_PUT_CHAR:	// 0x6c:
	case STATIC_PUT_SHORT:	// 0x6d:
	{
		info_register2T	*registersP;

		registersP = assemblerP->info.register2P;
		hash_register(&(registersP->reg1));
		hash_field(registersP->reg2.u.fieldP);
		break;
	}
	case INSTANCE_OF:		// 0x20: 22c -> B|A|op CCCC -> op vA, vB, type@CCCC
	case NEW_ARRAY:			// 0x23:
	{
		info_register3T	*registersP;

		registersP = assemblerP->info.register3P;
		hash_register(&(registersP->reg1));
		hash_register(&(registersP->reg2));
		hash_type(registersP->reg3.u.typeP);
		break;
	}
	case IGET:				// 0x52: 22c -> B|A|op CCCC -> op vA, vB, field@CCCC
	case IGET_WIDE:			// 0x53:
	case IGET_OBJECT:		// 0x54:
	case IGET_BOOLEAN:		// 0x55:
	case IGET_BYTE:			// 0x56:
	case IGET_CHAR:			// 0x57:
	case IGET_SHORT:		// 0x58:
	case IPUT:				// 0x59:
	case IPUT_WIDE:			// 0x5a:
	case IPUT_OBJECT:		// 0x5b:
	case IPUT_BOOLEAN:		// 0x5c:
	case IPUT_BYTE:			// 0x5d:
	case IPUT_CHAR:			// 0x5e:
	case IPUT_SHORT:		// 0x5f:
	{
		info_register3T	*registersP;

		registersP = assemblerP->info.register3P;
		hash_register(&(registersP->reg1));
		hash_register(&(registersP->reg2));
		hash_field(registersP->reg3.u.fieldP);
		break;
	}
	case GOTO_8:
	case GOTO_16:
		assert(0);
	case GOTO_32:
		if (assemblerP->m_jump.P < assemblerP) {
			++g_hash;
		}
		break;
	case IF_EQ_ZERO:		// 0x38: 21t -> AA|op BBBB -> op vAA, +BBBB
	case IF_NE_ZERO:		// 0x39:
	case IF_LT_ZERO:		// 0x3a:
	case IF_GE_ZERO:		// 0x3b:
	case IF_GT_ZERO:		// 0x3c:
	case IF_LE_ZERO:		// 0x3d:
	{
		info_register1T	*registersP;

		registersP = assemblerP->info.register1P;
		hash_register(&(registersP->reg1));
		if (assemblerP->m_jump.P < assemblerP) {
			++g_hash;
		}
		break;
	}
	case IF_EQ:				// 0x32: 22t -> B|A|op CCCC -> op vA, vB, +CCCC
	case IF_NE:				// 0x33:
	case IF_LT:				// 0x34:
	case IF_GE:				// 0x35:
	case IF_GT:				// 0x36:
	case IF_LE:				// 0x37:
	{
		info_register2T	*registersP;

		registersP = assemblerP->info.register2P;
		hash_register(&(registersP->reg1));
		hash_register(&(registersP->reg2));
		if (assemblerP->m_jump.P < assemblerP) {
			++g_hash;
		}
		break;
	}
	case FILL_ARRAY_DATA:	// 0x26: 31t -> AA|op BBBBlo BBBBhi -> op vAA, +BBBBBBBB fill-array-data vAA, +BBBBBBBB (with supplemental data as specified below in "fill-array-data Format")
	{
		info_fillT	*registersP;

		registersP = assemblerP->info.fillP;

		hash_register(&(registersP->reg1));
		hash_integer(registersP->width);
		hash_integer(registersP->entries);
		hash_data(registersP->width * registersP->entries, registersP->data);
		break;
	}
	case PACKED_SWITCH:
		assert(0);
	case SPARSE_SWITCH:
	{
		info_valueT	*registersP;

		registersP = assemblerP->info.valueP;

		hash_register(&(registersP->reg1));
		hash_integer(registersP->value);
		if (assemblerP->m_jump.P < assemblerP) {
			++g_hash;
		}
		break;
	}
	case FILLED_NEW_ARRAY:	// 0x25: 3rc -> AA|op BBBB CCCC -> op {vCCCC .. vNNNN}, type@BBBB (NNNN = AA + CCCC - 1}
	case FILL_ARRAY:		// 0x24: 35c -> B|A|op CCCC G|F|E|D -> [B=5] op {vD, vE, vF, vG, vA}, type@CCCC
	{
		info_registersT	*registersP;
		int				i;

		registersP = assemblerP->info.registersP;

		hash_type(registersP->last.u.typeP);
		for (i = registersP->size; i; ) {
			hash_register(registersP->reg + --i);
		}
		break;
	}
	case INVOKE_VIRTUAL_RANGE: // 0x74: 3rc -> AA|op BBBB CCCC -> op {vCCCC .. vNNNN}, meth@BBBB
	case INVOKE_SUPER_RANGE://0x75:
	case INVOKE_DIRECT_RANGE://0x76:
	case INVOKE_STATIC_RANGE://0x77:
	case INVOKE_INTERFACE_RANGE://0x78:
	case INVOKE_VIRTUAL:	// 0x6e: [B=5] op {vD, vE, vF, vG, vA}, meth@CCCC
	case INVOKE_SUPER:		// 0x6f: [B=4] op {vD, vE, vF, vG}, meth@CCCC
	case INVOKE_DIRECT:		// 0x70: [B=3] op {vD, vE, vF}, meth@CCCC
	case INVOKE_STATIC:		// 0x71: [B=2] op {vD, vE}, meth@CCCC
	case INVOKE_INTERFACE:	// 0x72: [B=1] op {vD}, meth@CCCC
	{
		info_registersT	*registersP;
		int				i;

		registersP = assemblerP->info.registersP;

		hash_method(registersP->last.u.methodP);
		for (i = registersP->size; i; ) {
			hash_register(registersP->reg + --i);
		}
		break;
	}
	case TRY_START:
	{
		info_tryT	*registersP = assemblerP->info.tryP;
		stringT		*typeP      = registersP->classtype.u.typeP;

		if (typeP) {
			hash_type(typeP);
		}
		if (assemblerP->m_jump.P < assemblerP) {
			++g_hash;
		}
		break;
	}
	case TRY_END:
	{
		info_tryT	*registersP = assemblerP->info.tryP;
		stringT		*typeP      = registersP->classtype.u.typeP;

		if (typeP) {
			hash_type(typeP);
		}
		break;
	}
	case BAD_3E:			// 0x3e: Bad operation
	case BAD_3F:			// 0x3f:
	case BAD_40:			// 0x40:
	case BAD_41:			// 0x41:
	case BAD_42:			// 0x42:
	case BAD_43:			// 0x43:
	case BAD_73:			// 0x73:
	case BAD_79:			// 0x79:
	case BAD_7A:			// 0x7a:
	case BAD_E3:			// 0xe3:
	case BAD_E4:			// 0xe4:
	case BAD_E5:			// 0xe5:
	case BAD_E6:			// 0xe6:
	case BAD_E7:			// 0xe7:
	case BAD_E8:			// 0xe8:
	case BAD_E9:			// 0xe9:
	case BAD_EA:			// 0xea:
	case BAD_EB:			// 0xeb:
	case BAD_EC:			// 0xec:
	case BAD_ED:			// 0xed:
	case BAD_EE:			// 0xee:
	case BAD_EF:			// 0xef:
	case BAD_F0:			// 0xf0:
	case BAD_F1:			// 0xf1:
	case BAD_F2:			// 0xf2:
	case BAD_F3:			// 0xf3:
	case BAD_F4:			// 0xf4:
	case BAD_F5:			// 0xf5:
	case BAD_F6:			// 0xf6:
	case BAD_F7:			// 0xf7:
	case BAD_F8:			// 0xf8:
	case BAD_F9:			// 0xf9:
	case BAD_FA:			// 0xfa:
	case BAD_FB:			// 0xfb:
	case BAD_FC:			// 0xfc:
	case BAD_FD:			// 0xfd:
		assert(0);
	}

	g_hash &= (HASH_SIZE - 1);
	assert(g_hash >= 0 && g_hash < HASH_SIZE);

	if ((hashPP = hash_table[g_hash])) {
		// Add to the end of this hash chain
		*hashPP = assemblerP; 
	}
	// Remember where end of this hash chain now is
	hash_table[g_hash] = &(assemblerP->m_hashP);
}
